'''
.. Copyright (c) 2016 Richard Plevin

   See the https://opensource.org/licenses/MIT for license details.
'''
#
#  Facilities setting up / customizing GCAM project's XML files.
#
# Common variables and functions for manipulating XML files.
# Basic approach is to create a directory for each defined scenario,
# in which modified files and a corresponding configuration XML file
# are stored.
#
# To allow functions to be called in any order or combination, each
# copies (if needed) the source file to the local scenario dir, then
# edits it in place. If was previously modified by another function,
# the copy is skipped, and the new edits are applied to the local,
# already modified file. Each function updates the local config file
# to refer to the modified file. (This may be done multiple times, to
# no ill effect.)
#
import glob
import os
import re
import shutil
import six
from lxml import etree as ET

from .config import getParam, getParamAsBoolean
from .constants import LOCAL_XML_NAME, DYN_XML_NAME, GCAM_32_REGIONS
from .error import SetupException, PygcamException
from .log import getLogger
from .utils import (coercible, mkdirs, unixPath, pathjoin, printSeries, symlinkOrCopyFile,
                    removeTreeSafely, parse_version_info)

# Names of key scenario components in reference GCAM 4.3 configuration.xml file
ENERGY_TRANSFORMATION_TAG = "energy_transformation"
SOLVER_TAG = "solver"

AttributePattern = re.compile('(.*)/@([-\w]*)$')
XmlDirPattern    = re.compile('/[^/]*-xml/')

_logger = getLogger(__name__)

# methods callable from <function name="x">args</function> in
# XML scenario setup scripts.
CallableMethods = {}

# decorator it identify callable methods
def callableMethod(func):
    CallableMethods[func.__name__] = func
    return func

def getCallableMethod(name):
    return CallableMethods.get(name)

def makeDirPath(elements, require=False, create=False, mode=0o775):
    """
    Join the tuple of elements to create a path to a directory,
    optionally checking that it exists or creating intermediate
    directories as needed.

    :param elements: a tuple of pathname elements to join
    :param require: if True, raise an error if the path doesn't exist
    :param create: if True, create the path if it doesn't exist
    :param mode: file mode used when making directories
    :return: the joined path
    :raises: pygcam.error.SetupException
    """
    path = pathjoin(*elements)

    if (create or require) and not os.path.lexists(path):
        if create:
            os.makedirs(path, mode)
        elif require:
            raise SetupException("Required path '%s' does not exist." % path)

    return path

#
# Copy src file to dst only if it doesn't already exist.
#
def copyIfMissing(src, dst, makedirs=False):
    """
    Copy file `src` to `dst`, but only if `dst` doesn't already exist.

    :param src: (str) pathname of the file to copy
    :param dst: (str) pathname of the copy to create
    :param makedirs: if True, make any missing directories
    :return: none
    """
    if not os.path.lexists(dst):
        parentDir = os.path.dirname(dst)
        if makedirs and not os.path.isdir(parentDir):
            _logger.debug("mkdir %s" % parentDir)
            os.makedirs(parentDir, 0o755)

        _logger.info("Copy %s\n      to %s" % (src, dst))
        shutil.copy(src, dst)
        os.chmod(dst, 0o644)

class CachedFile(object):
    parser = ET.XMLParser(remove_blank_text=True)

    # Store parsed XML trees here and use with xmlSel/xmlEdit if useCache is True
    cache = {}

    def __init__(self, filename):
        self.filename = filename
        self.edited = False

        _logger.debug("Reading '%s'", filename)
        self.tree = ET.parse(filename, self.parser)
        self.cache[filename] = self

    @classmethod
    def getFile(cls, filename):
        if filename in cls.cache:
            #_logger.debug("Found '%s' in cache", filename)
            item = cls.cache[filename]
        else:
            item = CachedFile(filename)

        return item

    def setEdited(self):
        self.edited = True

    def write(self):
        _logger.info("Writing '%s'", self.filename)
        self.tree.write(self.filename, xml_declaration=True, encoding='utf-8', pretty_print=True)
        self.edited = False

    def decache(self):
        if self.edited:
            self.write()

    @classmethod
    def decacheAll(cls):
        for item in cls.cache.values():
            item.decache()


def xmlSel(filename, xpath, asText=False):
    """
    Return True if the XML component identified by the xpath argument
    exists in `filename`. Useful for deciding whether to edit or
    insert an XML element.

    :param filename: (str) the file to edit
    :param xpath: (str) the xml element(s) to search for
    :param asText: (str) if True, return the text of the node, if found, else None
    :return: (bool) True if found, False otherwise. (see asText)
    """
    item = CachedFile.getFile(filename)
    result = item.tree.find(xpath)

    if asText:
        return result.text if result is not None else None

    return bool(result)


def xmlEdit(filename, pairs, useCache=True):
    """
    Edit the XML file `filename` in place, applying the values to the given xpaths
    in the list of pairs.

    :param filename: the file to edit in-place.
    :param pairs: (iterable of (xpath, value) pairs) In each pair, the xpath selects
      elements or attributes to update with the given values.
    :param useCache: (bool) if True, the etree is sought first in the XmlCache. This
      avoids repeated parsing, but the file is always written (eventually) if updated
      by this function.
    :return: True on success, else False
    """
    item = CachedFile.getFile(filename)
    tree = item.tree

    updated = False

    # if at least one xpath is found, update and write file
    for xpath, value in pairs:
        attr = None
        value = str(value)

        # If it's an attribute update, extract the attribute
        # and use the rest of the xpath to select the elements.
        match = re.match(AttributePattern, xpath)
        if match:
            attr = match.group(2)
            xpath = match.group(1)

        elts = tree.xpath(xpath)
        if len(elts):
            updated = True
            if attr:                # conditional outside loop since there may be many elements
                for elt in elts:
                    elt.set(attr, value)
            else:
                for elt in elts:
                    elt.text = value

    if updated:
        if useCache:
            item.setEdited()
        else:
            item.write()

    return updated

def extractStubTechnology(region, srcFile, dstFile, sector, subsector, technology,
                          sectorElement='supplysector', fromRegion=False):
    """
    Extract a definition from the global-technology-database based on `sector`, `subsector`,
    and `technology`, defined in `srcFile` and create a new file, `dstFile` with the extracted
    bit as a stub-technology definition for the given region. If `fromRegion` is True,
    extract the stub-technology from the regional definition, rather than from the
    global-technology-database.

    :param region: (str) the name of the GCAM region for which to copy the technology
    :param srcFile: (str) the pathname of a source XML file with a global-technology-database
    :param dstFile: (str) the pathname of the file to create
    :param sector: (str) the name of a GCAM sector
    :param subsector: (str) the name of a GCAM subsector within `sector`
    :param technology: (str) the name of a GCAM technology within `sector` and `subsector`
    :param sectorElement: (str) the name of the XML element to create (or search for, if `fromRegion`
        is True) between the ``<region>`` and ``<subsector>`` XML elements. Defaults to 'supplysector'.
    :param fromRegion: (bool) if True, the definition is extracted from a regional definition
        rather than from the global-technology-database.
    :return: True on success, else False
    """
    _logger.info("Extract stub-technology for %s (%s) to %s" % (technology, region if fromRegion else 'global', dstFile))

    if fromRegion:
        xpath = "//region[@name='%s']/%s[@name='%s']/subsector[@name='%s']/stub-technology[@name='%s']" % \
                (region, sectorElement, sector, subsector, technology)
    else:
        xpath = "//global-technology-database/location-info[@sector-name='%s' and @subsector-name='%s']/technology[@name='%s']" % \
                (sector, subsector, technology)

    # Read the srcFile to extract the required elements
    parser = ET.XMLParser(remove_blank_text=True)
    tree = ET.parse(srcFile, parser)

    # Rename technology => stub-technology (for global-tech-db case)
    elts = tree.xpath(xpath)
    if len(elts) != 1:
        raise PygcamException('Xpath "%s" failed' % xpath)

    technologyElt = elts[0]
    technologyElt.tag = 'stub-technology'       # no-op if fromRegion == True

    # Surround the extracted XML with the necessary hierarchy
    scenarioElt  = ET.Element('scenario')
    worldElt     = ET.SubElement(scenarioElt, 'world')
    regionElt    = ET.SubElement(worldElt, 'region', attrib={'name' : region})
    sectorElt    = ET.SubElement(regionElt, sectorElement, attrib={'name' : sector})
    subsectorElt = ET.SubElement(sectorElt, 'subsector', attrib={'name' : subsector})
    subsectorElt.append(technologyElt)

    # Workaround for parsing error: explicitly name shutdown deciders
    elts = scenarioElt.xpath("//phased-shutdown-decider|profit-shutdown-decider")
    for elt in elts:
        parent = elt.getparent()
        parent.remove(elt)

    _logger.info("Writing '%s'", dstFile)
    newTree = ET.ElementTree(scenarioElt)
    newTree.write(dstFile, xml_declaration=True, pretty_print=True)

    return True

def expandYearRanges(seq):
    """
    Expand a sequence of (year, value) tuples, or a dict keyed by
    year, where the year argument may be a string containing identifying
    range of values with an optional "step" value indicated after a ":".
    The default step is 5 years. For example, "2015-2030" expands to
    (2015, 2020, 2025, 2030), and "2015-2020:1" expands to
    (2015, 2016, 2017, 2018, 2019, 2020). When a range is given, the
    tuple is replaced with a sequence of tuples naming each year explicitly.
    Typical usage is ``for year, price in expandYearRanges(values): ...``.

    :param seq_or_dict:
        The sequence of (year, value) tuples, or any object with an
        iteritems() method that returns (year, value) pairs.
    :return:
        A list of tuples with the expanded sequence.
    """
    result = []
    try:
        seq = list(seq.items())     # convert dict or Series to list of pairs
    except:                             # of quietly fail, and just use 'seq' as is
        pass

    for year, value in seq:
        value = float(value)
        if isinstance(year, six.string_types) and '-' in year:
            m = re.search('^(\d{4})-(\d{4})(:(\d+))?$', year)
            if not m:
                raise SetupException('Unrecognized year range specification: %s' % year)

            startYear = int(m.group(1))
            endYear   = int(m.group(2))
            stepStr = m.group(4)
            step = int(stepStr) if stepStr else 5
            expanded = map(lambda y: [y, value], range(startYear, endYear+step, step))
            result.extend(expanded)
        else:
            result.append((int(year), value))

    return result

# TBD: maybe xmlSetup should be the only approach rather than supporting original setup subclasses.
# TBD: this way we can assume scenario definition exists in xml format and create an API to get the
# TBD: information about any scenario definition from xmlSetup.py.
#
# TBD: The question is whether command-line override capability is required, or if all should be in XML.
# TBD: Need to think through alternative use cases.
#
# TBD: should be no need to pass baseline since this can be inferred from scenario and scenarioGroup.
# TBD: also can tell if it's a baseline; if not, find and cache ref to baseline
class ScenarioInfo(object):
    def __init__(self, scenarioGroup, scenarioName, scenarioSubdir,
                 xmlSourceDir, xmlGroupSubdir, sandboxRoot, sandboxGroupSubdir):

        self.scenarioGroup  = scenarioGroup
        self.scenarioName   = scenarioName
        self.scenarioSubdir = scenarioSubdir or scenarioName

        self.xmlSourceDir = xmlSourceDir
        self.xmlGroupSubdir = xmlGroupSubdir or scenarioGroup

        self.sandboxRoot  = sandboxRoot
        self.sandboxGroupSubdir = sandboxGroupSubdir or scenarioGroup

        self.isBaseline = False # TBD

        if not self.isBaseline:
            self.baselineName = 'something'
            self.baselineInfo = self.fromXmlSetup(scenarioGroup, self.baselineName)

        # TBD: after setting self.x for all x:
        self.configPath = pathjoin(self.scenarioDir(), 'config.xml', realPath=True)

    @classmethod
    def fromXmlSetup(cls, scenarioGroup, scenarioName):
        # TBD: lookup the group and scenario, grab all data and
        # TBD: return ScenarioInfo(...)
        pass

    def absPath(self, x):
        pass

    def relPath(self, y):
        pass

    def scenarioXmlSourceDir(self):
        return pathjoin(self.xmlSourceDir, self.xmlGroupSubdir, self.scenarioSubdir, 'xml') # TBD if keeping xml subdir

    def scenarioXmlOutputDir(self):
        return pathjoin(self.xmlOutputDir, self.scenarioGroup, self.scenarioName)

    def scenarioXmlSourceFiles(self):
        files = glob.glob(self.scenarioXmlSourceDir() + '/*.xml')
        return files

    def cfgPath(self):
        """
        Compute the name of the GCAM config file for the current scenario.

        :return: (str) the pathname to the XML configuration file.
        """
        if not self.configPath:
            # compute the first time, then cache it
            self.configPath = unixPath(os.path.realpath(pathjoin(self.scenario_dir_abs, 'config.xml')))

        return self.configPath

class XMLEditor(object):
    '''
    Base class for scenario setup. Custom scenario processing classes must
    subclass this. Represents the information required to setup a scenario, i.e.,
    to generate and/or copy the required XML files into the XML output dir.
    '''
    # TBD: consider whether init should take an object describing the scenario
    # TBD: that can be populated from a scenario instance from xmlSetup.py or something
    # TBD: specific to the task. All these args are a pain, and there's no method API
    # TBD: to perform common ops.
    # TBD:
    def __init__(self, baseline, scenario, xmlOutputRoot, xmlSourceDir, refWorkspace,
                 groupDir, srcGroupDir, subdir, parent=None):
        self.name = name = scenario or baseline # if no scenario stated, assume baseline
        self.baseline = baseline
        self.scenario = scenario
        self.xmlOutputRoot = xmlOutputRoot
        self.refWorkspace = refWorkspace
        self.xmlSourceDir = xmlSourceDir
        self.sandboxExeDir = pathjoin(getParam('GCAM.SandboxRefWorkspace'), 'exe')
        self.parent = parent
        self.mcsMode = None
        self.mcsValues = None

        self.setupArgs = None

        # TBD: this would be just ../local-xml "project/scenario" occurs once, above
        # Allow scenario name to have arbitrary subdirs between "../local-xml" and
        # the scenario name, e.g., "../local-xml/project/scenario"
        self.subdir = subdir or ''
        self.groupDir = groupDir
        self.srcGroupDir = srcGroupDir or groupDir

        self.configPath = None

        # TBD: xmlOutputRoot is now just scenario dir, so this parameter can disappear
        self.local_xml_abs = makeDirPath((xmlOutputRoot, LOCAL_XML_NAME), create=True)
        self.dyn_xml_abs   = makeDirPath((xmlOutputRoot, DYN_XML_NAME), create=True) # TBD eliminate

        self.local_xml_rel = pathjoin("..", LOCAL_XML_NAME)
        self.dyn_xml_rel   = pathjoin("..", DYN_XML_NAME)   # TBD eliminate

        self.trial_xml_rel = self.trial_xml_abs = None      # used by MCS only

        # TBD: order changes using ScenarioInfo API
        self.scenario_dir_abs = makeDirPath((self.local_xml_abs, groupDir, name), create=True)
        self.scenario_dir_rel = pathjoin(self.local_xml_rel, groupDir, name)

        # Get baseline from ScenarioGroup and use ScenarioInfo API to get this type of info
        self.baseline_dir_rel = pathjoin(self.local_xml_rel, groupDir, self.parent.name) if self.parent else None

        # TBD eliminate
        self.scenario_dyn_dir_abs = makeDirPath((self.dyn_xml_abs, groupDir, name), create=True)
        self.scenario_dyn_dir_rel = pathjoin(self.dyn_xml_rel, groupDir, name)

        # Store commonly-used paths
        gcam_xml = "input/gcam-data-system/xml"
        self.gcam_prefix_abs = prefix_abs = pathjoin(refWorkspace, gcam_xml)
        self.gcam_prefix_rel = prefix_rel = pathjoin('../', gcam_xml)

        # TBD: maybe no need to store these since computable from rel paths
        self.aglu_dir_abs           = pathjoin(prefix_abs, 'aglu-xml')
        self.emissions_dir_abs      = pathjoin(prefix_abs, 'emissions-xml')
        self.energy_dir_abs         = pathjoin(prefix_abs, 'energy-xml')
        self.modeltime_dir_abs      = pathjoin(prefix_abs, 'modeltime-xml')
        self.socioeconomics_dir_abs = pathjoin(prefix_abs, 'socioeconomics-xml')

        self.aglu_dir_rel           = pathjoin(prefix_rel, 'aglu-xml')
        self.emissions_dir_rel      = pathjoin(prefix_rel, 'emissions-xml')
        self.energy_dir_rel         = pathjoin(prefix_rel, 'energy-xml')
        self.modeltime_dir_rel      = pathjoin(prefix_rel, 'modeltime-xml')
        self.socioeconomics_dir_rel = pathjoin(prefix_rel, 'socioeconomics-xml')

        # TBD: add climate and policy subdirs?
        self.solution_prefix_abs = pathjoin(refWorkspace, "input", "solution")
        self.solution_prefix_rel = pathjoin("..", "input", "solution")

    def absPath(self, relPath):
        """
        Convert `relPath` to an absolute path by treating it as relative to
        the current scenario's "exe" directory.

        :param relPath: (str) a path relative to the current "exe" directory
        :return: (str) the absolute path corresponding to `relPath`.
        """
        return pathjoin(self.xmlOutputRoot, 'exe', relPath, normpath=True)

    @staticmethod
    def recreateDir(path):
        removeTreeSafely(path)
        mkdirs(path)

    def setupDynamic(self, args):
        """
        Create dynamic XML files in dyn-xml. These files are generated for policy
        scenarios when XML file contents must be computed from baseline results.

        :param args: (argparse.Namespace) arguments passed from the top-level call
            to setup sub-command
        :return: none
        """

        _logger.info("Generating dyn-xml for scenario %s" % self.name)

        # Delete old generated scenario files
        dynDir = self.scenario_dyn_dir_abs
        self.recreateDir(dynDir)

        scenDir = self.scenario_dir_abs
        xmlFiles = glob.glob("%s/*.xml" % scenDir)

        # TBD: no need to link or copy if all in one place. [But dyn are per-trial; local are not]
        if xmlFiles:
            mode = 'Copy' if getParamAsBoolean('GCAM.CopyAllFiles') else 'Link'
            _logger.info("%s %d static XML files in %s to %s", mode, len(xmlFiles), scenDir, dynDir)

            for xml in xmlFiles:
                base = os.path.basename(xml)
                dst = pathjoin(dynDir, base)
                src = pathjoin(scenDir, base)
                symlinkOrCopyFile(src, dst)
        else:
            _logger.info("No XML files to link in %s", unixPath(os.path.abspath(scenDir)))

        CachedFile.decacheAll()

    def setupStatic(self, args):
        """
        Create static XML files in local-xml. By "static", we mean files whose contents are
        independent of baseline results. In comparison, policy scenarios may generate dynamic
        XML files whose contents are computed from baseline results.

        :param args: (argparse.Namespace) arguments passed from the top-level call to setup
            sub-command.
        :return: none
        """
        _logger.info("Generating local-xml for scenario %s" % self.name)

        scenDir = self.scenario_dir_abs
        mkdirs(scenDir)

        # TBD: there's nothing else now in these dirs, so "xml" subdir is not needed
        xmlSubdir = pathjoin(self.xmlSourceDir, self.srcGroupDir, self.subdir or self.name, 'xml')
        xmlFiles = glob.glob("%s/*.xml" % xmlSubdir)

        if xmlFiles:
            _logger.info("Copy %d static XML files from %s to %s", len(xmlFiles), xmlSubdir, scenDir)
            for src in xmlFiles:
                # dst = pathjoin(scenDir, os.path.basename(src))
                shutil.copy2(src, scenDir)     # copy2 preserves metadata, e.g., timestamp
        else:
            _logger.info("No XML files to copy in %s", unixPath(os.path.abspath(xmlSubdir)))

        configPath = self.cfgPath()

        parent = self.parent
        parentConfigPath = parent.cfgPath() if parent else getParam('GCAM.RefConfigFile')

        _logger.info("Copy %s\n      to %s" % (parentConfigPath, configPath))
        shutil.copy(parentConfigPath, configPath)
        os.chmod(configPath, 0o664)

        # set the scenario name
        self.updateConfigComponent('Strings', 'scenarioName', self.name)

        # This is inherited from baseline by policy scenarios; no need to redo this
        if not self.parent:
            self.makeScenarioComponentsUnique()

        # For the following configuration file settings, no action is taken when value is None
        if args.stopPeriod is not None:
            self.setStopPeriod(args.stopPeriod)

        # For the following boolean arguments, we first check if there is any value. If
        # not, no change is made. If a value is given, the parameter is set accordingly.
        if getParam('GCAM.WritePrices'):
            self.updateConfigComponent('Bools', 'PrintPrices', int(getParamAsBoolean('GCAM.WritePrices')))

        if getParam('GCAM.WriteDebugFile'):
            self.updateConfigComponent('Files', 'xmlDebugFileName', value=None,
                                       writeOutput=getParamAsBoolean('GCAM.WriteDebugFile'))

        if getParam('GCAM.WriteXmlOutputFile'):
            self.updateConfigComponent('Files', 'xmlOutputFileName', value=None,
                                       writeOutput=getParamAsBoolean('GCAM.WriteXmlOutputFile'))

        # According to Pralit, outFile.csv isn't maintained and isn't reliable.
        if getParam('GCAM.WriteOutputCsv'):
            self.updateConfigComponent('Files', 'outFileName', value=None,
                                       writeOutput=getParamAsBoolean('GCAM.WriteOutputCsv'))

        CachedFile.decacheAll()

    def setup(self, args):
        """
        Calls setupStatic and/or setupDynamic, depending on flags set in args.

        :param args: (argparse.Namespace) arguments passed from the top-level call
            to setup
        :return: none
        """
        _logger.debug('Called XMLEditor.setup(%s)', args)
        self.setupArgs = args   # some subclasses/functions might want access to these

        if not args.dynamicOnly:
            self.setupStatic(args)

        if not args.staticOnly:
            self.setupDynamic(args)

        CachedFile.decacheAll()

    def makeScenarioComponentsUnique(self):
        """
        Give all reference ScenarioComponents a unique "name" tag to facilitate
        manipulation via xmlstarlet. This is a no-op in GCAM v4.3.

        :return: none
        """
        version = parse_version_info()

        # no longer necessary in 4.3. For 4.2, we reset names to those used in 4.3
        if version < (4, 3):
            self.renameScenarioComponent("interest_rate", pathjoin(self.socioeconomics_dir_rel, "interest_rate.xml"))
            self.renameScenarioComponent("socioeconomics", pathjoin(self.socioeconomics_dir_rel, "socioeconomics_GCAM3.xml"))

            self.renameScenarioComponent("industry", pathjoin(self.energy_dir_rel, "industry.xml"))
            self.renameScenarioComponent("industry_income_elas", pathjoin(self.energy_dir_rel, "industry_incelas_gcam3.xml"))

            self.renameScenarioComponent("cement", pathjoin(self.energy_dir_rel, "cement.xml"))
            self.renameScenarioComponent("cement_income_elas", pathjoin(self.energy_dir_rel, "cement_incelas_gcam3.xml"))

            self.renameScenarioComponent("fertilizer_energy", pathjoin(self.energy_dir_rel, "en_Fert.xml"))
            self.renameScenarioComponent("fertilizer_agriculture", pathjoin(self.aglu_dir_rel, "ag_Fert.xml"))

            for i in (1,2,3):
                tag = 'land%d' % i
                filename = 'land_input_%d.xml' % i
                self.renameScenarioComponent(tag, pathjoin(self.aglu_dir_rel, filename))

                if i > 1:
                    tag = 'protected_' + tag
                    filename = 'protected_' + filename
                    self.renameScenarioComponent(tag, pathjoin(self.aglu_dir_rel, filename))

    def cfgPath(self):
        """
        Compute the name of the GCAM config file for the current scenario.

        :return: (str) the pathname to the XML configuration file.
        """
        if not self.configPath:
            # compute the first time, then cache it
            self.configPath = unixPath(os.path.realpath(pathjoin(self.scenario_dir_abs, 'config.xml')))

        return self.configPath

    def componentPath(self, tag, configPath=None):
        configPath = configPath or self.cfgPath()
        pathname = xmlSel(configPath, '//Value[@name="%s"]' % tag, asText=True)

        if pathname is None:
            raise PygcamException("Failed to find scenario component with tag '%s' in %s" % (tag, configPath))

        return pathname

    def getLocalCopy(self, configTag):
        """
        Get the filename for the most local version (in terms of scenario hierarchy)
        of the XML file identified in the configuration file with `configTag`, and
        copy the file to our scenario dir if not already there.

        :param configTag: (str) the configuration file tag (name="xxx") of an XML file
        :return: (str, str) a tuple of the relative and absolute path of the
          local (i.e., within the current scenario) copy of the file.
        """
        # if configTag.endswith('.xml'):
            # It's not a tag, but a filename

        pathname = self.componentPath(configTag)
        srcAbsPath = os.path.abspath(os.path.join(self.sandboxExeDir, pathname))

        # TBD: test this
        if not os.path.lexists(srcAbsPath):
            _logger.debug("Didn't find %s; checking reference files" % srcAbsPath)
            # look to sandbox workspace if not found locally
            refWorkspace  = getParam('GCAM.SandboxRefWorkspace')
            refConfigFile = getParam('GCAM.RefConfigFile')

            pathname = self.componentPath(configTag, configPath=refConfigFile)
            srcAbsPath = os.path.abspath(os.path.join(refWorkspace, 'exe', pathname))

        # If path includes /*-xml/* (e.g., '/energy-xml/', '/aglu-xml/'), retain
        # this subdir in destination, else just use the basename of the path.
        matches = list(re.finditer(XmlDirPattern, srcAbsPath))
        if matches:
            m = matches[-1]
            suffix = srcAbsPath[m.start()+1:]   # from after '/' to end
        else:
            suffix = os.path.basename(srcAbsPath)

        dstAbsPath = pathjoin(self.scenario_dir_abs, suffix)
        dstRelPath = pathjoin(self.scenario_dir_rel, suffix)

        copyIfMissing(srcAbsPath, dstAbsPath, makedirs=True)

        return dstRelPath, dstAbsPath

    def updateConfigComponent(self, group, name, value=None, writeOutput=None, appendScenarioName=None):
        """
        Update the value of an arbitrary element in GCAM's configuration.xml file, i.e.,
        ``<{group}><Value name="{name}>{value}</Value></{group}>``

        Optional args are used only for ``<Files>`` group, which has entries like
        ``<Value write-output="1" append-scenario-name="0" name="outFileName">outFile.csv</Value>``
        Values for the optional args can be passed as any of ``[0, 1, "0", "1", True, False]``.

        :param group: (str) the name of a group of config elements in GCAM's configuration.xml
        :param name: (str) the name of the element to be updated
        :param value: (str) the value to set between the ``<Value></Value>`` elements
        :param writeOutput: (coercible to int) for ``<Files>`` group, this sets the optional ``write-output``
           attribute
        :param appendScenarioName: (coercible to int) for ``<Files>`` group, this sets the optional
          ``append-scenario-name`` attribute.
        :return: none
        """
        textArgs = "name='%s'" % name
        if writeOutput is not None:
            textArgs += " write-output='%d'" % (int(writeOutput))
        if appendScenarioName is not None:
            textArgs += " append-scenario-name='%d'" % (int(appendScenarioName))

        _logger.debug("Update <%s><Value %s>%s</Value>" % (group, textArgs, '...' if value is None else value))

        cfg = self.cfgPath()

        prefix = "//%s/Value[@name='%s']" % (group, name)
        pairs = []

        if value is not None:
            pairs.append((prefix, value))

        if writeOutput is not None:
            pairs.append((prefix + "/@write-output", int(writeOutput)))

        if appendScenarioName is not None:
            pairs.append((prefix + "/@append-scenario-name", int(appendScenarioName)))

        xmlEdit(cfg, pairs)

    @callableMethod
    def setClimateOutputInterval(self, years):
        """
        Sets the the frequency at which climate-related outputs are
        saved to the XML database to the given number of years,
        e.g., ``<Value name="climateOutputInterval">1</Value>``.
        **Callable from XML setup files.**

        :param years: (coercible to int) the number of years to set as the climate (GHG)
           output interval
        :return: none
        """
        self.updateConfigComponent('Ints', 'climateOutputInterval', coercible(years, int))

    def addScenarioComponent(self, name, xmlfile):
        """
        Add a new ``<ScenarioComponent>`` to the configuration file, at the end of the list
        of components.

        :param name: (str) the name to assign to the new scenario component
        :param xmlfile: (str) the location of the XML file, relative to the `exe` directory
        :return: none
        """
        # Ensure no duplicates tags
        self.deleteScenarioComponent(name)

        xmlfile = unixPath(xmlfile)
        _logger.info("Add ScenarioComponent name='%s', xmlfile='%s'" % (name, xmlfile))

        cfg = self.cfgPath()
        item = CachedFile.getFile(cfg)
        item.setEdited()

        elt = item.tree.find('//ScenarioComponents')
        node = ET.SubElement(elt, 'Value')
        node.set('name', name)
        node.text = xmlfile

    def insertScenarioComponent(self, name, xmlfile, after):
        """
        Insert a ``<ScenarioComponent>`` to the configuration file, following the
        entry named by ``after``.

        :param name: (str) the name to assign to the new scenario component
        :param xmlfile: (str) the location of the XML file, relative to the `exe` directory
        :param after: (str) the name of the element after which to insert the new component
        :return: none
        """
        # Ensure no duplicates tags
        self.deleteScenarioComponent(name)

        xmlfile = unixPath(xmlfile)
        _logger.info("Insert ScenarioComponent name='%s', xmlfile='%s' after value '%s'" % (name, xmlfile, after))

        cfg = self.cfgPath()
        item = CachedFile.getFile(cfg)
        item.setEdited()

        elt = item.tree.find('//ScenarioComponents')
        afterNode = elt.find('Value[@name="%s"]' % after)
        if afterNode is None:
            raise SetupException("Can't insert %s after %s, as the latter doesn't exist" % (name, after))

        index = elt.index(afterNode) + 1

        node = ET.Element('Value')
        node.set('name', name)
        node.text = xmlfile
        elt.insert(index, node)

    def updateScenarioComponent(self, name, xmlfile):
        """
        Set a new filename for a ScenarioComponent identified by the ``<Value>`` element name.

        :param name: (str) the name of the scenario component to update
        :param xmlfile: (str) the location of the XML file, relative to the `exe` directory, that
           should replace the existing value
        :return: none
        """
        xmlfile = unixPath(xmlfile)

        self.updateConfigComponent('ScenarioComponents', name, xmlfile)

    def deleteScenarioComponent(self, name, useCache=True):
        """
        Delete a ``<ScenarioComponent>`` identified by the ``<Value>`` element name.

        :param name: (str) the name of the ScenarioComponent to delete
        :return: none
        """
        _logger.info("Delete ScenarioComponent name='%s' for scenario" % name)
        cfg = self.cfgPath()
        item = CachedFile.getFile(cfg)

        elt = item.tree.find("//ScenarioComponents")
        valueNode = elt.find("Value[@name='%s']" % name)
        if valueNode is not None:
            elt.remove(valueNode)
            item.setEdited()

    def renameScenarioComponent(self, name, xmlfile):
        """
        Modify the name of a ``ScenarioComponent``, located by the XML file path it holds.
        This is used in to create a local reference XML that has unique names
        for all scenario components, which allows all further modifications to refer
        only to the (now unique) names.

        :param name: (str) the new name for the scenario component
        :param xmlfile: (str) the XML file path used to locate the scenario component
        :return: none
        """
        xmlfile = unixPath(xmlfile)
        _logger.debug("Rename ScenarioComponent name='%s', xmlfile='%s'" % (name, xmlfile))
        cfg = self.cfgPath()

        xmlEdit(cfg, ("//ScenarioComponents/Value[text()='%s']/@name" % xmlfile, name))

    # TBD dynamic keyword might still be useful if subdir e.g. local-xml/dynamic but policy file would be in local-xml anyway
    @callableMethod
    def addMarketConstraint(self, target, policy, dynamic=False,
                            baselinePolicy=False): # TBD: should be able to eliminate this arg
        """
        Adds references to a pair of files comprising a policy, i.e., a policy definition
        file and a constraint file. References to the two files--assumed to be named ``XXX-{subsidy,tax}.xml``
        and ``XXX-{subsidy,tax}-constraint.xml`` for policy `target` ``XXX``--are added to the configuration file.
        **Callable from XML setup files.**

        :param target: (str) the subject of the policy, e.g., corn-etoh, cell-etoh, ft-biofuel, biodiesel
        :param policy: (str) one of ``subsidy`` or ``tax``
        :param dynamic: (str) True if the XML file was dynamically generated, and thus found in ``dyn-xml``
           rather than ``local-xml``
        :param baselinePolicy: (bool) if True, the policy file is linked to the baseline directory
           rather than this scenario's own directory.
        :return: none
        """
        _logger.info("Add market constraint: %s %s for %s" % (target, policy, self.name))

        cfg = self.cfgPath()

        basename = "%s-%s" % (target, policy)	# e.g., biodiesel-subsidy

        policyTag     = target + "-policy"
        constraintTag = target + "-constraint"

        reldir = self.scenario_dyn_dir_rel if dynamic else self.scenario_dir_rel

        # TBD: Could look for file in scenario, but if not found, look in baseline, eliminating this flag
        policyReldir = self.baseline_dir_rel if baselinePolicy else reldir

        policyXML     = pathjoin(policyReldir, basename + ".xml") # TBD: "-market.xml" for symmetry?
        constraintXML = pathjoin(reldir, basename + "-constraint.xml")

        # See if element exists in config file (-Q => quiet; just report exit status)
        xpath = '//ScenarioComponents/Value[@name="%s"]' % policyTag

        # If we've already added files for policy/constraint on this target,
        # we replace the old values with new ones. Otherwise, we add them.
        addOrUpdate = self.updateScenarioComponent if xmlSel(cfg, xpath) else self.addScenarioComponent
        addOrUpdate(policyTag, policyXML)
        addOrUpdate(constraintTag, constraintXML)

    @callableMethod
    def delMarketConstraint(self, target, policy):
        """
        Delete the two elements defining a market constraint from the configuration file.
        The filenames are constructed as indicated in the `addMarketConstraint` method.
        **Callable from XML setup files.**

        :param target: (str) the subject of the policy, e.g., corn-etoh, cell-etoh,
            ft-biofuel, biodiesel
        :param policy: (str) one of ``subsidy`` or ``tax``
        :return: none
        """
        _logger.info("Delete market constraint: %s %s for %s" % (target, policy, self.name))
        cfg = self.cfgPath()

        # if policy == "subsidy":
        #     policy = "subs"	# use shorthand in filename

        policyTag     = target + "-" + policy
        constraintTag = target + "-constraint"

        # See if element exists in config file (-Q => quiet; just report exit status)
        xpath = '//ScenarioComponents/Value[@name="%s"]' % policyTag

        if xmlSel(cfg, xpath):
            # found it; delete the elements
            self.deleteScenarioComponent(policyTag)
            self.deleteScenarioComponent(constraintTag)

    @callableMethod
    def setStopPeriod(self, yearOrPeriod):
        """
        Sets the model stop period. If `stopPeriod` is <= 22, the stop period is set to
        the given value. If the value > 2000, the value is treated as a year and converted
        to the correct stop period for the configuration file.
        **Callable from XML setup files.**

        :param yearOrPeriod: (coercible to int) this argument is treated as a literal
          stop period if the value is < 1000. (N.B. 2015 = step 4, 2020 = step 5, and so
          on.) If yearOrPeriod >= 1000, it is treated as a year and converted
          to a stopPeriod for use in the GCAM configuration file.
        :return: none
        :raises: SetupException
        """
        value = coercible(yearOrPeriod, int)
        stopPeriod = value if 1 < value < 1000 else 1 + (value - 2000)/5

        self.updateConfigComponent('Ints', 'stop-period', stopPeriod)

    @callableMethod
    def setInterpolationFunction(self, region, supplysector, subsector, fromYear, toYear,
                                 funcName='linear', applyTo='share-weight', stubTechnology=None,
                                 delete=False):
        """
        Set the interpolation function for the share-weight of the `subsector`
        of `supplysector` to `funcName` between years `fromYear` to `toYear`
        in `region`. **Callable from XML setup files.**

        :param region: (str) the GCAM region to operate on
        :param supplysector: (str) the name of a supply sector
        :param subsector: (str) the name of a sub-sector
        :param fromYear: (str or int) the year to start interpolating
        :param toYear: (str or int) the year to stop interpolating
        :param funcName: (str) the name of an interpolation function
        :param applyTo: (str) what the interpolation function is applied to
        :param stubTechnology: (str) the name of a technology to apply function to
        :param delete: (bool) if True, set delete="1", otherwise don't.
        :return: none
        """
        _logger.info("Set interpolation function for '%s' : '%s' to '%s'" % (supplysector, subsector, funcName))

        enTransFileRel, enTransFileAbs = self.getLocalCopy(ENERGY_TRANSFORMATION_TAG)

        # /scenario/world/region[@name='USA']/supplysector[@name='refining']/subsector[@name='biomass liquids']/interpolation-rule
        prefix = '//region[@name="%s"]/supplysector[@name="%s"]/subsector[@name="%s"]%s/interpolation-rule[@apply-to="%s"]' % \
                 (region, supplysector, subsector,
                  '/stub-technology[@name="%s"]' % stubTechnology if stubTechnology else '',
                  applyTo)

        args = [(prefix + '/@from-year', str(fromYear)),
                (prefix + '/@to-year', str(toYear)),
                (prefix + '/interpolation-function/@name', funcName)]

        if delete:
            args.append((prefix + '/@delete', "1"))

        xmlEdit(enTransFileAbs, args)

        self.updateScenarioComponent("energy_transformation", enTransFileRel)

    @callableMethod
    def setupSolver(self, solutionTolerance=None, broydenTolerance=None,
                    maxModelCalcs=None, maxIterations=None):
        """
        Set the model solution tolerance to the given values for the solver
        "driver" (`solutionTolerance`) and, optionally for the Broyden component
        (`broydenTolerance`).
        **Callable from XML setup files.**

        :param solutionTolerance: (coercible to float, > 0.0) the value to set for the driver tolerance
        :param broydenTolerance: (coercible to float, > 0.0) the value to set for the Broyden component
            tolerance. (If both are provided, the function requires that
            componentTolerance <= driverTolerance.)
        :param maxModelCalcs: (coercible to int, > 0) maximum number of calculations to run in the driver
        :param maxIterations: (coercible to int, > 0) maximum number of iterations to allow in the
            Broyden component
        :return: none
        """
        def coercibleAndPositive(name, value, requiredType):
            if value is None:
                return None

            value = coercible(value, requiredType)
            if value <= 0:
                raise SetupException(name + ' must be greater than zero')

            _logger.info("Set %s to %s", name, value)
            return value

        solutionTol = coercibleAndPositive('Driver solution tolerance', solutionTolerance, float)
        broydenTol  = coercibleAndPositive('Broyden component tolerance', broydenTolerance, float)

        if solutionTol and broydenTol:
            if broydenTol > solutionTol:
                raise SetupException('Broyden component tolerance cannot be greater than driver solution tolerance')

        maxModelCalcs = coercibleAndPositive('maxModelCalcs', maxModelCalcs, int)
        maxIterations = coercibleAndPositive('maxIterations', maxIterations, int)

        solverFile = 'cal_broyden_config.xml'
        solverFileRel, solverFileAbs = self.getLocalCopy(SOLVER_TAG)

        prefix = "//scenario/user-configurable-solver[@year>=2010]/"
        pairs = []

        if solutionTolerance:
            pairs.append((prefix + 'solution-tolerance', solutionTolerance))

        if broydenTolerance:
            pairs.append((prefix + 'broyden-solver-component/ftol', broydenTolerance))

        if maxModelCalcs:
            pairs.append((prefix + 'max-model-calcs', maxModelCalcs))

        if maxIterations:
            pairs.append((prefix + 'broyden-solver-component/max-iterations', maxIterations))

        xmlEdit(solverFileAbs, pairs)

        self.updateScenarioComponent("solver", solverFileRel)

    @callableMethod
    def dropLandProtection(self, dropEmissions=True):
        self.deleteScenarioComponent("protected_land2")
        self.deleteScenarioComponent("protected_land3")

        if dropEmissions:
            version = parse_version_info()
            if version > (5, 0):
                # drop emissions for protected land
                self.deleteScenarioComponent("nonco2_aglu_prot")

    @callableMethod
    def protectLand(self, fraction, landClasses=None, otherArable=False,
                    regions=None, unprotectFirst=False):
        """
        Modify land_input files to protect a constant fraction of unmanaged
        land of the given classes, in the given regions.
        **Callable from XML setup files.**

        :param fraction: (float) the fraction of land in the given land classes
               to protect
        :param landClasses: a string or a list of strings, or None. If None, all
               "standard" unmanaged land classes are modified.
        :param otherArable: (bool) if True, land class 'OtherArableLand' is
            included in default land classes.
        :param regions: a string or a list of strings, or None. If None, all
               regions are modified.
        """
        from .landProtection import protectLand

        _logger.info("Protecting %d%% of land globally", int(fraction * 100))

        # NB: this code depends on these being the tags assigned to the land files
        # as is currently the case in XmlEditor.makeScenarioComponentsUnique()
        for num in [2, 3]:
            fileTag  = 'land%d' % num
            landFileRel, landFileAbs = self.getLocalCopy(fileTag)

            protectLand(landFileAbs, landFileAbs, fraction, landClasses=landClasses,
                        otherArable=otherArable, regions=regions, unprotectFirst=unprotectFirst)
            self.updateScenarioComponent(fileTag, landFileRel)

    # TBD: test
    @callableMethod
    def protectionScenario(self, scenarioName, unprotectFirst=True):
        """
        Implement the protection scenario `scenarioName`, defined in the file given
        by config variable `GCAM.LandProtectionXmlFile`.
        **Callable from XML setup files.**

        :param scenarioName: (str) the name of a scenario defined in the land
           protection XML file.
        :param unprotectFirst: (bool) if True, make all land "unprotected" before
           protecting.
        :return: none
        """
        from .landProtection import runProtectionScenario

        _logger.info("Using protection scenario %s", scenarioName)

        landXmlFiles = []

        # NB: this code depends on these being the tags assigned to the land files
        # as is currently the case in XmlEditor.makeScenarioComponentsUnique()
        for num in [2, 3]:
            fileTag  = 'land%d' % num
            landFileRel, landFileAbs = self.getLocalCopy(fileTag)

            landXmlFiles.append(landFileAbs)
            self.updateScenarioComponent(fileTag, landFileRel)

        # TBD: revisit this; it's a bit of a hack for Oct 16 deliverable
        scenarioFile = pathname = getParam('GCAM.LandProtectionXmlFile')
        if self.mcsMode == 'trial':
            basename = os.path.basename(pathname)
            scenario = self.scenario or self.baseline
            scenarioFile = unixPath(pathjoin(self.trial_xml_abs, 'local-xml',
                                             self.groupDir, scenario, basename))

        runProtectionScenario(scenarioName, scenarioFile=scenarioFile, inPlace=True,
                              xmlFiles=landXmlFiles, unprotectFirst=unprotectFirst)

    def getScenarioOrTrialDirs(self, subdir=''):
        dirRel = pathjoin(self.trial_xml_rel, subdir) if self.mcsMode == 'trial' \
            else self.scenario_dir_rel

        dirAbs = pathjoin(self.trial_xml_abs, subdir) if self.mcsMode == 'trial' \
            else self.scenario_dir_abs

        return dirRel, dirAbs

    @callableMethod
    def taxCarbon(self, value, startYear=2020, endYear=2100, timestep=5,
                  rate=0.05, regions=GCAM_32_REGIONS, market='global'):
        '''
        Generate an XML file defining a global carbon tax starting
        at `value` and increasing by `rate` annually. Generate values
        for the give `years`. The first year in `years` is assumed to be
        the year at which the tax starts at `value`. The generated file
        is named 'carbon-tax-{market}.xml' and is added to the configuration.
        **Callable from XML setup files.**

        :param value: (float) the initial value of the tax ($/tonne)
        :param years: (list(int)) years to set carbon taxes. Default is 2020-2100
           at 10 year time-steps.
        :param rate: (float) annual rate of increase. Default is 0.05.
        :param regions: (list(str)) the regions for which to create a C tax market.
             Default is all 32 GCAM regions.
        :param market: (str) the name of the market to create. Default is 'global'.
        :return: none
        '''
        from .carbonTax import genCarbonTaxFile

        tag = 'carbon-tax-' + market
        filename = tag + '.xml'

        # TBD: need to generalize this since any modification can be per-trial or universal
        dirRel, dirAbs = self.getScenarioOrTrialDirs(subdir='local-xml')

        fileRel = pathjoin(dirRel, filename)
        fileAbs = pathjoin(dirAbs, filename)

        genCarbonTaxFile(fileAbs, value, startYear=startYear, endYear=endYear,
                         timestep=timestep, rate=rate, regions=regions, market=market)
        self.addScenarioComponent(tag, fileRel)

    @callableMethod
    def taxBioCarbon(self, market='global', regions=None, forTax=True, forCap=False):
        """
        Create the XML for a linked policy to include LUC CO2 in a CO2 cap or tax policy (or both).
        This function generates the equivalent of any of the 4 files in input/policy/:
        global_ffict.xml               (forTax=False, forCap=False)
        global_ffict_in_constraint.xml (forTax=False, forCap=True)
        global_uct.xml                 (forTax=True,  forCap=False)
        global_uct_in_constraint.xml   (forTax=True,  forCap=True)

        However, unlike those files, the market need not be global, and the set of regions to
        which to apply the policy can be specified.

        :param market: (str) the name of the market for which to create the linked policy
        :param regions: (list of str or None) the regions to apply the policy to, or None
          to indicate all regions.
        :param forTax: (bool) True if the linked policy should apply to a CO2 tax
        :param forCap: (bool) True if the linked policy should apply to a CO2 cap
        :return: (str) the generated XML text
        """
        from .carbonTax import genLinkedBioCarbonPolicyFile

        tag = 'bio-carbon-tax-' + market
        filename = tag + '.xml'

        # TBD: need to generalize this since any modification can be per-trial or universal
        dirRel, dirAbs = self.getScenarioOrTrialDirs(subdir='local-xml')

        fileRel = pathjoin(dirRel, filename)
        fileAbs = pathjoin(dirAbs, filename)

        genLinkedBioCarbonPolicyFile(fileAbs, market=market, regions=regions,
                                     forTax=forTax, forCap=forCap)
        self.addScenarioComponent(tag, fileRel)

    # TBD: test
    @callableMethod
    def setRegionPopulation(self, region, values):
        """
        Set the population for the given region to the values for the given years.
        **Callable from XML setup files.**

        :param region: (str) the name of one of GCAM's regions.
        :param values: (dict-like or iterable of tuples of (year, pop)), specifying
           the population to set for each year given.
        :return: none
        """
        # msg = "Set population for %s in %s to:" % (region, self.name)
        # printSeries(values, region, header=msg, loglevel='INFO')

        tag = 'socioeconomics'
        #path = self.componentPath(tag)
        # fileRel, fileAbs = self.getLocalCopy(path)
        fileRel, fileAbs = self.getLocalCopy(tag)

        prefix = '//region[@name="%s"]/demographics/populationMiniCAM' % region
        pairs = []
        for year, pop in expandYearRanges(values):
            pairs.append((prefix + ('[@year="%s"]/totalPop' % year), int(round(pop))))

        xmlEdit(fileAbs, pairs)
        self.updateScenarioComponent(tag, fileRel)

    # TBD: test
    @callableMethod
    def setGlobalTechNonEnergyCost(self, sector, subsector, technology, values):
        """
        Set the non-energy cost of for technology in the global-technology-database,
        given a list of values of (year, price). The price is applied to all years
        indicated by the range.
        **Callable from XML setup files.**

        :param sector: (str) the name of a GCAM sector
        :param subsector: (str) the name of a GCAM subsector within `sector`
        :param technology: (str) the name of a GCAM technology in `subsector`
        :param values: (dict-like or iterable of tuples of (year, price)) `year` can
            be a single year (as string or int), or a string specifying a range of
            years, of the form "xxxx-yyyy", which implies 5 year timestep, or "xxxx-yyyy:s",
            which provides an alternative timestep. If `values` is dict-like (e.g. a
            pandas Series) a list of tuples is created by calling values.iteritems() after
            which the rest of the explanation above applies. The `price` can be
            anything coercible to float.
        :return: none
        """
        msg = "Set non-energy-cost of %s for %s to:" % (technology, self.name)
        _logger.info(printSeries(values, technology, header=msg, asStr=True))

        #_logger.info("Set non-energy-cost of %s for %s to %s" % (technology, self.name, values))

        enTransFileRel, enTransFileAbs = self.getLocalCopy(ENERGY_TRANSFORMATION_TAG)

        prefix = '//global-technology-database/location-info[@sector-name="%s" and @subsector-name="%s"]/technology[@name="%s"]' % \
                 (sector, subsector, technology)
        suffix = '/minicam-non-energy-input[@name="non-energy"]/input-cost'

        pairs = []
        for year, price in expandYearRanges(values):
            pairs.append((prefix + ('/period[@year="%s"]' % year) + suffix, price))

        xmlEdit(enTransFileAbs, pairs)

        self.updateScenarioComponent("energy_transformation", enTransFileRel)

    # TBD: Test
    @callableMethod
    def setGlobalTechShutdownRate(self, sector, subsector, technology, values):
        """
        Create a modified version of en_transformation.xml with the given shutdown
        rates for `technology` in `sector` based on the data in `values`.
        **Callable from XML setup files.**

        :param sector: (str) the name of a GCAM sector
        :param subsector: (str) the name of a GCAM subsector within `sector`
        :param technology: (str) the name of a GCAM technology in `subsector`
        :param values: (dict-like or iterable of tuples of (year, shutdownRate)) `year` can
            be a single year (as string or int), or a string specifying a range of
            years, of the form "xxxx-yyyy", which implies 5 year timestep, or "xxxx-yyyy:s",
            which provides an alternative timestep. If `values` is dict-like (e.g. a
            pandas Series) a list of tuples is created by calling values.iteritems() after
            which the rest of the explanation above applies. The `shutdownRate` can be
            anything coercible to float.
        :param xmlBasename: (str) the name of an xml file in the energy-xml folder to edit.
        :param configFileTag: (str) the 'name' of a <File> element in the <ScenarioComponents>
           section of a config file. This must match `xmlBasename`.
        :return: none
        """
        _logger.info("Set shutdown rate for (%s, %s) to %s for %s" % (sector, technology, values, self.name))

        enTransFileRel, enTransFileAbs = self.getLocalCopy(ENERGY_TRANSFORMATION_TAG)

        prefix = "//global-technology-database/location-info[@sector-name='%s' and @subsector-name='%s']/technology[@name='%s']" % \
                 (sector, subsector, technology)

        pairs = []

        for year, value in expandYearRanges(values):
            pairs.append((prefix + "/period[@year='%s']/phased-shutdown-decider/shutdown-rate" % year,
                         coercible(value, float)))

        xmlEdit(enTransFileAbs, pairs)
        self.updateScenarioComponent("energy_transformation", enTransFileRel)

    #
    # //region[@name=""]/energy-final-demand[@name=""]/price-elasticity[@year=""]
    #
    # names of energy-final-demand:
    # 'aglu-xml/demand_input.xml': "Exports_Meat", "FoodDemand_Crops", "FoodDemand_Meat", "NonFoodDemand_Crops", "NonFoodDemand_Forest", "NonFoodDemand_Meat"
    # 'energy-xml/transportation_UCD.xml': "trn_aviation_intl", "trn_freight", "trn_pass", "trn_shipping_intl"
    # 'energy-xml/cement.xml: "cement"
    # 'energy-xml/industry.xml: "industry"
    #
    @callableMethod
    def setPriceElasticity(self, regions, sectors, configFileTag, values):
        """
        Modify price-elasticity values for the given `regions` and `sectors` in `sector` based on the data in `values`.
        **Callable from XML setup files.**

        :param regions: (str or list of str) the name(s) of a GCAM region or regions, or "global"
           to indicate that price elasticity should be set in all regions. (Or more precisely,
           the change should not be restricted by region.)
        :param sector: (str or list of str) the name of a GCAM (demand) sector. In GCAM v4.3, this
            should be one of {"cement", "industry", "trn_aviation_intl", "trn_freight", "trn_pass",
            "trn_shipping_intl", "Exports_Meat", "FoodDemand_Crops", "FoodDemand_Meat",
            "NonFoodDemand_Crops", "NonFoodDemand_Forest", "NonFoodDemand_Meat"}, however if input
            files have been customized, other values can be used.
        :param configFileTag: (str) the 'name' of a <File> element in the <ScenarioComponents>
           section of a config file. This determines which file is edited, so it must correspond to
           the indicated sector(s).
        :param values: (dict-like or iterable of tuples of (year, elasticity)) `year` can
            be a single year (as string or int), or a string specifying a range of
            years, of the form "xxxx-yyyy", which implies 5 year timestep, or "xxxx-yyyy:s",
            which provides an alternative timestep. If `values` is dict-like (e.g. a
            pandas Series) a list of tuples is created by calling values.iteritems() after
            which the rest of the explanation above applies. The `elasticity` can be
            anything coercible to float.
        :return: none
        """
        _logger.info("Set price-elasticity for (%s, %s) to %s for %s" % (regions, sectors, values, self.name))

        filenameRel, filenameAbs = self.getLocalCopy(configFileTag)

        def listifyString(value, aliasForNone=None):
            if isinstance(value, six.string_types):
                value = [value]

            # Treat "global" as not restricting by region
            if aliasForNone and len(value) == 1 and value[0] == aliasForNone:
                return None

            return value

        def nameExpression(values):
            '''
            Turn ['a', 'b'] into '@name="a" or @name="b"'
            '''
            names = ['@name="%s"' % v for v in values]
            return ' or '.join(names)

        regions = listifyString(regions, aliasForNone='global')
        nameExpr = '[' + nameExpression(regions) + ']' if regions else ''
        regionExpr = '//region' + nameExpr

        prefix = regionExpr + '/energy-final-demand[%s]' % nameExpression(sectors)

        pairs = []
        for year, value in expandYearRanges(values):
            pairs.append((prefix + '/price-elasticity[@year="%s"]' % year, coercible(value, float)))

        xmlEdit(filenameAbs, pairs)
        self.updateScenarioComponent(configFileTag, filenameRel)

    # TBD: test
    @callableMethod
    def setRegionalShareWeights(self, region, sector, subsector, values,
                               stubTechnology=None,
                               configFileTag=ENERGY_TRANSFORMATION_TAG):
        """
        Create a modified version of en_transformation.xml with the given share-weights
        for `technology` in `sector` based on the data in `values`.
        **Callable from XML setup files.**

        :param region: if not None, changes are made in a specific region, otherwise they're
            made in the global-technology-database.
        :param sector: (str) the name of a GCAM sector
        :param technology: (str) the name of a GCAM technology in `sector`
        :param values: (dict-like or iterable of tuples of (year, shareWeight)) `year` can
            be a single year (as string or int), or a string specifying a range of
            years, of the form "xxxx-yyyy", which implies 5 year timestep, or "xxxx-yyyy:s",
            which provides an alternative timestep. If `values` is dict-like (e.g. a
            pandas Series) a list of tuples is created by calling values.iteritems() after
            which the rest of the explanation above applies. The `shareWeight` can be
            anything coercible to float.
        :param configFileTag: (str) the 'name' of a <File> element in the <ScenarioComponents>
           section of a config file. This must match `xmlBasename`.
        :return: none
        """
        from .utils import printSeries

        _logger.info("Set share-weights for (%r, %r, %r, %r) for %r",
                     region, sector, subsector, stubTechnology, self.name)
        _logger.info(printSeries(values, 'share-weights', asStr=True))

        enTransFileRel, enTransFileAbs = self.getLocalCopy(configFileTag)

        prefix = "//region[@name='%s']/supplysector[@name='%s']/subsector[@name='%s']" % (region, sector, subsector)

        shareWeight = '/stub-technology[@name="{technology}"]/period[@year="{year}"]/share-weight' \
                      if stubTechnology else '/share-weight[@year="{year}"]'

        pairs = []
        for year, value in expandYearRanges(values):
            pairs.append((prefix + shareWeight.format(technology=stubTechnology, year=year),
                         coercible(value, float)))

        xmlEdit(enTransFileAbs, pairs)
        self.updateScenarioComponent(configFileTag, enTransFileRel)

    # TBD: Test
    @callableMethod
    def setGlobalTechShareWeight(self, sector, subsector, technology, values,
                                 configFileTag=ENERGY_TRANSFORMATION_TAG):
        """
        Create a modified version of en_transformation.xml with the given share-weights
        for `technology` in `sector` based on the data in `values`.
        **Callable from XML setup files.**

        :param sector: (str) the name of a GCAM sector
        :param technology: (str) the name of a GCAM technology in `sector`
        :param values: (dict-like or iterable of tuples of (year, shareWeight)) `year` can
            be a single year (as string or int), or a string specifying a range of
            years, of the form "xxxx-yyyy", which implies 5 year timestep, or "xxxx-yyyy:s",
            which provides an alternative timestep. If `values` is dict-like (e.g. a
            pandas Series) a list of tuples is created by calling values.iteritems() after
            which the rest of the explanation above applies. The `shareWeight` can be
            anything coercible to float.
        :param xmlBasename: (str) the name of an xml file in the energy-xml folder to edit.
        :param configFileTag: (str) the 'name' of a <File> element in the <ScenarioComponents>
           section of a config file. This must match `xmlBasename`.
        :return: none
        """
        _logger.info("Set share-weights for (%s, %s) to %s for %s" % (sector, technology, values, self.name))

        enTransFileRel, enTransFileAbs = self.getLocalCopy(configFileTag)

        prefix = "//global-technology-database/location-info[@sector-name='%s' and @subsector-name='%s']/technology[@name='%s']" % \
                 (sector, subsector, technology)

        pairs = []
        for year, value in expandYearRanges(values):
            pairs.append((prefix + "/period[@year=%s]/share-weight" % year, coercible(value, float)))

        xmlEdit(enTransFileAbs, pairs)
        self.updateScenarioComponent(configFileTag, enTransFileRel)

    # TBD: test
    @callableMethod
    def setEnergyTechnologyCoefficients(self, subsector, technology, energyInput, values):
        '''
        Set the coefficients in the global technology database for the given energy input
        of the given technology in the given subsector.
        **Callable from XML setup files.**

        :param subsector: (str) the name of the subsector
        :param technology: (str)
            The name of the technology, e.g., 'cellulosic ethanol', 'FT biofuel', etc.
        :param energyInput: (str) the name of the minicam-energy-input
        :param values:
            A sequence of tuples or object with ``iteritems`` method returning
            (year, coefficient). For example, to set
            the coefficients for cellulosic ethanol for years 2020 and 2025 to 1.234,
            the pairs would be ((2020, 1.234), (2025, 1.234)).
        :return:
            none
        '''
        _logger.info("Set coefficients for %s in global technology %s, subsector %s: %s" % \
                     (energyInput, technology, subsector, values))

        enTransFileRel, enTransFileAbs = \
            self.getLocalCopy(ENERGY_TRANSFORMATION_TAG)

        prefix = "//global-technology-database/location-info[@subsector-name='%s']/technology[@name='%s']" % \
                 (subsector, technology)
        suffix = "minicam-energy-input[@name='%s']/coefficient" % energyInput

        pairs = []
        for year, coef in expandYearRanges(values):
            pairs.append(("%s/period[@year='%s']/%s" % (prefix, year, suffix), coef))

        xmlEdit(enTransFileAbs, pairs)
        self.updateScenarioComponent("energy_transformation", enTransFileRel)



from __future__ import print_function
from six import iteritems
from itertools import chain
import ipyparallel as ipp
from ipyparallel.client.client import ExecuteReply


# Remotely-called function; imports requirement internally.
def dummyTask(key):
    from os import getpid
    from time import sleep
    from ipyparallel.datapub import publish_data

    publish_data({key: 'running'})
    sleep(2 * key)
    publish_data({key: 'finishing'})
    sleep(2)
    return {'key': key, 'pid': getpid()}


# Default time to wait in main hub polling/result processing loop
SLEEP_SECONDS = 3


class BaseMaster(object):
    instance = None

    @classmethod
    def getInstance(cls, profile=None, cluster_id=None, sleepSeconds=SLEEP_SECONDS):
        """
        Return singleton instance of this class
        """
        if not cls.instance:
            cls.instance = cls(profile=profile, cluster_id=cluster_id)

        cls.instance.setSleepSeconds(sleepSeconds)
        return cls.instance

    def __init__(self, profile=None, cluster_id=None):
        self.client = ipp.Client(profile=profile, cluster_id=cluster_id)
        self.statusDict = {}
        self.sleepSeconds = SLEEP_SECONDS
        self.keyField = 'key'

    def setSleepSeconds(self, secs):
        self.sleepSeconds = secs

    def clearStatus(self):
        self.statusDict = {}

    def setStatus(self, key, status):
        self.statusDict[key] = status

    def queueTotals(self):
        """
        Return totals for queue status across all engines
        """
        try:
            dv = self.client[:]
        except ipp.NoEnginesRegistered as e:
            print('queueTotals: %s' % e)
            return

        qstatus = dv.queue_status()

        totals = dict(queue=0, completed=0, tasks=0)

        for id, stats in iteritems(qstatus):
            if id == u'unassigned':
                totals[id] = stats
            else:
                for key, count in iteritems(stats):
                    totals[key] += count

        return totals

    def runningTasks(self):
        qstatus = self.client.queue_status(verbose=True)
        ids = [rec['tasks'] for key, rec in iteritems(qstatus) if isinstance(key, (int, long))]
        return list(chain.from_iterable(ids))

    def completedTasks(self):
        recs = self.client.db_query({'completed': {'$ne': None}}, keys=['msg_id'])
        ids = [rec['msg_id'] for rec in recs] if recs else None
        return ids

    def getResults(self, tasks):
        if not tasks:
            return None
        client = self.client
        ar = client.get_result(tasks, owner=True, block=False)
        try:
            results = ar.get()
        except Exception as e:
            print('getResults: %s' % e)
            return

        client.purge_results(jobs=tasks) # so we don't see them again

        # filter out results from execute commands (e.g. imports)
        results = [r[0] for r in results if r and not isinstance(r, ExecuteReply)]
        return results

    def runTasks(self, count, clearStatus=False):
        if clearStatus:
            self.clearStatus()

        view = self.client.load_balanced_view()
        arList = []

        for key in range(1, count + 1):
            ar = view.apply_async(dummyTask, key)
            arList.append(ar)
            self.setStatus(key, 'queued')

        return arList

    def checkRunning(self):
        running = self.runningTasks()
        if running:
            try:
                # _logger.debug("Found %d running tasks", len(running))
                ar = self.client.get_result(running, block=False)
                statusDict = self.statusDict
                # print('statusDict:', statusDict)
                for dataDict in ar.data:
                    for key, status in iteritems(dataDict):
                        currStatus = statusDict.get(key)
                        if currStatus != status:
                            self.setStatus(key, status)
            except Exception as e:
                print("checkRunning: %s" % e)
                return

    def processResult(self, result):
        key = result[self.keyField]
        self.setStatus(key, 'completed')
        print("Completed", key)

    def processResults(self):
        from time import sleep

        while True:
            sleep(self.sleepSeconds)
            self.checkRunning()

            tot = self.queueTotals()
            print(tot)
            if not (tot['queue'] or tot['tasks'] or tot['unassigned']) and \
                    not self.completedTasks():
                return

            completed = self.completedTasks()
            if completed:
                results = self.getResults(completed)
                if not results:
                    print("Completed tasks have no results: engine died?")
                    continue    # is this recoverable?

                for result in results:
                    self.processResult(result)


if __name__ == '__main__':

    #
    # Test with custom worker func and subclass
    #
    def runTrial(argDict):
        from time import sleep
        from random import random
        from ipyparallel.datapub import publish_data

        def randomSleep(minSleep, maxSleep):
            delay = minSleep + random() * (maxSleep - minSleep)
            sleep(delay)
            argDict['slept'] = '%.2f' % delay

        runId = argDict['runId']

        publish_data({runId: 'running'})
        randomSleep(10, 15)
        publish_data({runId: 'finishing'})
        sleep(2)
        return argDict


    class NewMaster(BaseMaster):
        def __init__(self, profile=None, cluster_id=None):
            super(NewMaster, self).__init__(profile=profile, cluster_id=cluster_id)
            self.keyField = 'runId'

        def runTrials(self, tuples, clearStatus=False):
            if clearStatus:
                self.clearStatus()

            view = self.client.load_balanced_view(retries=1)    # in case engine fails, retry job once only
            asyncResults = []
            argDict = {}

            try:
                for runId, trialNum in tuples:
                    argDict['trialNum'] = trialNum
                    argDict['runId'] = runId

                    # Easier to deal with a list of AsyncResults than a single
                    # instance that contains info about all "future" results.
                    result = view.map_async(runTrial, [argDict])
                    asyncResults.append(result)
                    self.setStatus(runId, 'queued')

            except Exception as e:
                print("Exception running 'runTrial': %s", e)

        def processResult(self, result):
            key = result[self.keyField]
            self.setStatus(key, 'completed')
            print("Completed", result)


    testBase = False
    profile  = None
    cluster_id = None

    if testBase:
        m = BaseMaster.getInstance(sleepSeconds=5)
        m.runTasks(6, clearStatus=True)
    else:
        m = NewMaster.getInstance(sleepSeconds=3, profile=profile, cluster_id=cluster_id)
        tuples = [(runId, trialNum) for runId, trialNum in enumerate(range(1000, 1020))]
        m.runTrials(tuples, clearStatus=True)

    m.processResults()
    print('Status:')
    d = m.statusDict
    for runId in sorted(d.keys()):
        print('  runId %s: %s' %(runId, d[runId]))



'''
.. codeauthor:: Richard Plevin

.. Copyright (c) 2017 Richard Plevin
   See the https://opensource.org/licenses/MIT for license details.
'''
from __future__ import print_function
import itertools
import os
from six.moves import input

from ..config import getHomeDir
from ..subcommand import SubcommandABC

class AbortInput(Exception):
    pass

DefaultProjectDir = '~/GCAM/projects'

def defaultSandboxDir(projectRoot):
    from ..utils import pathjoin
    path = pathjoin(os.path.dirname(projectRoot), 'sandboxes')
    return path

Template = '''[DEFAULT]
GCAM.DefaultProject  = {dfltProject}
GCAM.RefWorkspace    = {gcamDir}
GCAM.ProjectRoot     = {projectRoot}
GCAM.SandboxRoot     = {sandboxRoot}
GCAM.RewriteSetsFile = %(GCAM.ProjectDir)s/etc/rewriteSets.xml

[{dfltProject}]
# Set GCAM.LogLevel to DEBUG for more diagnostic messages 
# or to WARN, ERROR, or FATAL for progressively less info. 
GCAM.LogLevel = INFO

# Setup generated config files to not write large extraneous 
# output files. Move these statements to the [DEFAULT] section 
# if you want this to be the default for all projects.
GCAM.WriteDebugFile     = False
GCAM.WriteXmlOutputFile = False
'''

def expandTilde(path):
    """
    Similar to os.path.expanduser() but uses our getHomeDir() function,
    which produces consistent results regards of whether user is running
    a standard Windows command window or a Cygwin bash shell. Also, we
    have no need to expand '~user', just '~'.
    """
    from ..utils import unixPath

    if path.startswith('~/'):
        path = getHomeDir() + path[1:]
    return unixPath(path)

def findGCAM():
    import platform
    from ..utils import unixPath

    home = getHomeDir()

    versions = ['gcam-v4.4', 'gcam-v4.3']
    dirs = [home, home + '/GCAM', home + '/gcam']

    system = platform.system()
    if system == 'Darwin':
        release = 'gcam-v4.4-Mac-Release-Package'
    elif system == 'Windows':
        release = 'gcam-v4.4-Windows-Release-Package'
    else:
        release = None

    if release:
        versions.insert(0, release)

    for v, d in itertools.product(versions, dirs):
        path = '%s/%s' % (d, v)
        if os.path.isdir(path):
            return unixPath(path)

    return ''

def askYesNo(msg, default=None):
    default = default and default.lower()
    y = 'Y' if default == 'y' else 'y'
    n = 'N' if default == 'n' else 'n'
    prompt = msg + ' (%s/%s)? ' % (y, n)

    value = None
    while value is None:
        value = input(prompt).lower()
        if value == '' and default:
            return default == 'y'

        if value not in ('y', 'n', 'yes', 'no'):
            value = None

    return value in ('y', 'yes')


def askString(msg, default):
    value = None
    while value is None:
        value = input(msg + ' (default=%s): ' % default)
        value = value.strip()
        if value == '':
            value = default

    return value

def askDir(msg, default=''):
    from prompt_toolkit import prompt
    from prompt_toolkit.contrib.completers import PathCompleter
    from ..utils import mkdirs

    is_cygwin = isCygwin()

    completer = PathCompleter(only_directories=True, expanduser=True)
    msg     = unicode(msg) + u' '
    default = unicode(default)
    path = None

    while not path:
        # prompt_toolkit doesn't work in cygwin terminal
        path = askString(msg, default) if is_cygwin else prompt(msg, default=default, completer=completer)
        path = expandTilde(path)

        if path and not os.path.isdir(path):
            if os.path.lexists(path):
                print('Path %s exists, but is not a directory. Try again.' % path)
                path = ''
                continue

            create = askYesNo("Path %s does not exist. Create it" % path, default='y')
            if not create:
                path = ''
                continue

            mkdirs(path)
            print('Created %s' % path)

    return path

def isCygwin():
    import sys, platform

    return (platform.system() == 'Windows'
            and os.environ.get('SHELL') == '/bin/bash'
            and not sys.stdin.isatty())

def rerunForCygwin(args):
    """
    CygWin buffers stdin unless PYTHONUNBUFFERED is set, so we
    set it and re-run the 'gt init' command with the given args.
    A bit of an unfortunate work-around, but it works.
    """
    import subprocess as subp

    os.environ['PYTHONUNBUFFERED'] = '1'

    argList = ['gt', 'init']

    if args.createProject == True:
        argList.append('-c')
    elif args.createProject == False:
        argList.append('-C')

    if args.overwrite:
        argList.append('--overwrite')

    if args.gcamDir:
        argList += ['-g', args.gcamDir]

    if args.defaultProject:
        argList += ['-P', args.defaultProject]

    if args.projectDir:
        argList += ['-p', args.projectDir]

    if args.sandboxDir:
        argList += ['-s', args.sandboxDir]

    command = ' '.join(argList)
    # print('Re-running for cygwin: %s' % command)
    subp.call(argList, shell=False)


class InitCommand(SubcommandABC):
    def __init__(self, subparsers):
        kwargs = {'help' : '''Initialize key variables in the  ~/.pygcam.cfg
            configuration file. Values not provided on the command-line are
            requested interactively.'''}

        super(InitCommand, self).__init__('init', subparsers, kwargs, group='utils')

    def addArgs(self, parser):
        group = parser.add_mutually_exclusive_group()

        group.add_argument('-c', '--create-project',    dest='createProject', action='store_true',
                            default=None,
                            help='''Create the project structure for the given default project. If 
                            neither -c/--create-project nor -C/--no-create-project is specified, the
                            user is queried interactively.''')

        group.add_argument('-C', '--no-create-project', dest='noCreateProject', action='store_true',
                            default=None,
                           help='''Do not create the project structure for the given default project.
                           Mutually exclusive with -c / --create-project option.''')

        parser.add_argument('-g', '--gcamDir',
                            help='''The directory that is a GCAM v4.3 or v4.4
                            workspace. Sets config var GCAM.RefWorkspace. By default,
                            looks for gcam-v4.4 (then v4.3) in ~, ~/GCAM, and ~/gcam, 
                            where "~" indicates your home directory.''')

        parser.add_argument('--overwrite', action='store_true',
                            help='''Overwrite an existing config file. (Makes
                            a backup first in ~/.pygcam.cfg~.)''')

        parser.add_argument('-P', '--defaultProject',
                            help='''Set the value of config var GCAM.DefaultProject to
                                    the given value.''')

        parser.add_argument('-p', '--projectDir',
                            help='''The directory in which to create pygcam project
                            directories. Sets config var GCAM.ProjectRoot. Default
                            is "%s".''' % DefaultProjectDir)

        parser.add_argument('-s', '--sandboxDir',
                            help='''The directory in which to create pygcam project
                            directories. Sets config var GCAM.SandboxRoot. Default
                            is "%s".''' % defaultSandboxDir(DefaultProjectDir))

        return parser

    def run(self, args, tool):
        from ..config import USR_CONFIG_FILE
        from ..utils import pathjoin, deleteFile, mkdirs

        # Detect cygwin terminals, which quite lamely cannot handle interactive input
        if (isCygwin() and not os.environ.get('PYTHONUNBUFFERED')):
            rerunForCygwin(args)
            return

        configPath = pathjoin(getHomeDir(), USR_CONFIG_FILE)

        try:
            dfltProject = args.defaultProject or askString('Default project name?', 'tutorial')
            gcamDir     = args.gcamDir or askDir('Where is GCAM installed?', default=findGCAM())
            projectDir  = args.projectDir or askDir('Directory in which to create pygcam projects?',
                                                    default=expandTilde(DefaultProjectDir))
            sandboxDir  = args.sandboxDir or askDir('Directory in which to create pygcam run-time sandboxes?',
                                                    default=defaultSandboxDir(projectDir))

            # make backup of configuration file if it exists and is not zero length
            if os.path.lexists(configPath) and os.stat(configPath).st_size > 0:
                overwrite = args.overwrite or askYesNo('Overwrite %s' % configPath, default='n')
                if not overwrite:
                    raise AbortInput("Quitting to avoid overwriting %s" % configPath)

                backup = configPath + '~'
                deleteFile(backup)
                os.rename(configPath, backup)
                print('Moved %s to %s' % (configPath, backup))

        except AbortInput as e:
            print(e)
            return

        # initialize configuration file
        text = Template.format(dfltProject=dfltProject, gcamDir=gcamDir,
                               projectRoot=projectDir, sandboxRoot=sandboxDir)

        with open(configPath, 'w') as f:
            f.write(text)

        print("Created %s with contents:\n\n%s" % (configPath, text))

        for path in (projectDir, sandboxDir):
            if not os.path.isdir(path):
                mkdirs(path)

        # These args default to None so we can test for explicit settings
        createProj = True if args.createProject == True else (False if args.noCreateProject == True else None)
        if createProj is None:
            createProj = askYesNo('Create the project structure for "%s"' % dfltProject, default='y')

        if createProj:
            newProjectDir = pathjoin(projectDir, dfltProject)
            overwrite = askYesNo('Overwrite existing project dir %s' % newProjectDir, 'n') if os.path.lexists(newProjectDir) else False

            argList = ['new', dfltProject, '-r', projectDir] + (['--overwrite'] if overwrite else [])

            args = tool.parser.parse_args(args=argList)
            tool.run(args=args)
            print('Created project "%s" in %s' % (dfltProject, newProjectDir))

PluginClass = InitCommand



'''
.. Created on: 2/12/15
   Common functions and data

.. Copyright (c) 2015-2016 Richard Plevin
   See the https://opensource.org/licenses/MIT for license details.
'''
import argparse
import io
import os
from lxml import etree as ET
import pkgutil
import re
import semver
import shutil
import subprocess
import sys
from contextlib import contextmanager

from .config import getParam, getParamAsBoolean
from .error import PygcamException, FileFormatError
from .log import getLogger
from .windows import IsWindows

_logger = getLogger(__name__)


def parse_version_info(vers=None):
    vers = vers or getParam('GCAM.VersionNumber')

    # if only major.minor is given (e.g., "4.4"), add .patch of zero (e.g., "4.4.0")
    if re.match('^\d\.\d$', vers):
        vers += '.0'

    return semver.parse_version_info(vers)

#
# Custom argparse "action" to parse comma-delimited strings to lists
# TBD: Use this where relevant
#
class ParseCommaList(argparse.Action):
    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        if nargs is not None:
            raise ValueError("nargs not allowed with " % option_strings)

        super(ParseCommaList, self).__init__(option_strings, dest, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, values.split(','))

# For gcam 5, read the map of river basins to countries (and other info)
def readBasinMap():
    from .query import readCsv
    from .utils import resourceStream

    stream = resourceStream('etc/gcam-v5/basin_to_country_mapping.csv')
    df = readCsv(stream, skiprows=2)
    return df

# For gcam 5, produce a map of basins for each (ISO) country name
def basinsByISO():
    df = readBasinMap()
    ISOs = df['ISO'].unique()
    result = {}
    for iso in ISOs:
        rows = df.query("ISO == '{}'".format(iso))
        result[iso] = list(rows['GLU_name'])

    return result

def queueForStream(stream):
    """
    Create a thread to read from a non-socket file descriptor and
    its contents to a socket so non-blocking read via select() works
    on Windows. (Since Windows doesn't support select on pipes.)

    :param stream: (file object) the input to read from,
       presumably a pipe from a subprocess
    :return: (int) a file descriptor for the socket to read from.
    """
    from six.moves.queue import Queue
    from threading import Thread

    def enqueue(stream, queue):
        fd = stream.fileno()
        data = None
        while data != b'':
            data = os.read(fd, 1024)
            queue.put(data)
        stream.close()

    q = Queue()
    t = Thread(target=enqueue, args=(stream, q))
    t.daemon = True # thread dies with the program
    t.start()

    return q

# Used only in CI plugins
def digitColumns(df, asInt=False):
    '''
    Get a list of columns with integer names (as strings, e.g., "2007") in df.
    If asInt is True return as a list of integers, otherwise as strings.
    '''
    digitCols = filter(str.isdigit, df.columns)
    return map(int, digitCols) if asInt else digitCols

# Function to return current function name, or the caller, and so on up
# the stack, based on value of n.
getFuncName = lambda n=0: sys._getframe(n + 1).f_code.co_name

pat = re.compile('(\{[^\}]+\})')

# used only in project.py
def simpleFormat(s, varDict):
    """
    Simple version of str.format that does not treat '.' as
    an attribute reference.

    :param s: (str) string with args in curly braces
    :param varDict: (dict) dictionary of var names and values
    :return: (str) formatted string
    """
    def lookup(m):
        match = m.group(0)
        key = match[1:-1]   # strip off curly braces
        return str(varDict[key])

    try:
        result = re.sub(pat, lookup, s)
        return result
    except KeyError as e:
        raise FileFormatError('Unknown parameter %s in project XML template string' % e)

@contextmanager
def pushd(directory):
    """
    Context manager that changes to the given directory and then
    returns to the original directory. Usage is ``with pushd('/foo/bar'): ...``

    :param directory: (str) a directory to chdir to temporarily
    :return: none
    """
    owd = os.getcwd()
    try:
        os.chdir(directory)
        yield directory
    finally:
        os.chdir(owd)

def getResource(relpath):
    """
    Extract a resource (e.g., file) from the given relative path in
    the pygcam package.

    :param relpath: (str) a path relative to the pygcam package
    :return: the file contents
    """
    contents = pkgutil.get_data('pygcam', relpath)
    return contents

def resourceStream(relpath):
    """
    Return a stream on the resource found on the given path relative
    to the pygcam package.

    :param relpath: (str) a path relative to the pygcam package
    :return: (file-like stream) a file-like buffer opened on the desired resource.
    """
    text = getResource(relpath)
    return io.BytesIO(text)

def copyResource(relpath, dest, overwrite=True):
    """
    Copy a resource from the 'pygcam' package to the given destination.

    :param relpath: (str) a path relative to the pygcam package
    :param dest: (str) the pathname of the file to create by copying the resource.
    :param overwrite: (bool) if False, raise an error if the destination
      file already exists.
    :return: none
    """
    if not overwrite and os.path.lexists(dest):
        raise FileFormatError(dest)

    text = getResource(relpath)
    with open(dest, 'w') as f:
        f.write(text)

# used only in gcamtool modules
def getBooleanXML(value):
    """
    Get a value from an XML file and convert it into a boolean True or False.

    :param value: any value (it's first converted to a string)
    :return: True if the value is in ['true', 'yes', '1'], False if the value
             is in ['false', 'no', '0']. An exception is raised if any other
             value is passed.
    :raises: PygcamException
    """
    false = ["false", "no", "0"]
    true  = ["true", "yes", "1"]

    val = str(value).strip()
    if val not in true + false:
        raise PygcamException("Can't convert '%s' to boolean; must be in {false,no,0,true,yes,1} (case sensitive)." % value)

    return (val in true)


_XMLDBPropertiesTemplate = """<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE properties SYSTEM "http://java.sun.com/dtd/properties.dtd">
<!-- WARNING: this file is automatically generated. Manual edits will be overwritten. -->
<properties>
	<entry key="in-memory">{inMemory}</entry>
	<entry key="filter-script">{filterFile}</entry>
	<entry key="batch-logfile">{batchLog}</entry>
	<entry key="batch-queries">{batchFile}</entry>
</properties>
"""

def writeXmldbDriverProperties(outputDir='.', inMemory=True, filterFile='', batchFile='', batchLog=''):
    """
    Write a XMLDBDriver.properties file using the values passed in the arguments.

    :param outputDir: (str) where to write the file
    :param inMemory: (bool) if True, the ``in-memory`` attribute is set to True
    :param filterFile: (str) a file that filters GCAM query output to limit what's
       written to the database
    :param batchFile: (str) the path to an XML batch query file
    :param batchLog: (str) the path to a log file into which to direct
       batch query messages (queries can be pretty verbose...)
    :return: none
    """
    path = pathjoin(outputDir, 'XMLDBDriver.properties')
    memFlag = 'true' if inMemory else 'false'
    content = _XMLDBPropertiesTemplate.format(inMemory=memFlag, filterFile=filterFile,
                                              batchFile=batchFile, batchLog=batchLog)

    deleteFile(path) # avoid writing through symlinks to ref workspace
    with open(path, 'w') as f:
        f.write(content)

def deleteFile(filename):
    """
    Delete the given `filename`, but ignore errors, like "rm -f"

    :param filename: (str) the file to remove
    :return: none
    """
    try:
        os.remove(filename)
    except:
        pass    # ignore errors, like "rm -f"

# used only in gcamtool modules
def symlinkOrCopyFile(src, dst):
    """
    Symlink a file unless GCAM.CopyAllFiles is True, in which case, copy the file.

    :param src: (str) filename of original file
    :param dst: (dst) filename of copy
    :return: none
    """
    if getParamAsBoolean('GCAM.CopyAllFiles'):
        copyFileOrTree(src, dst)
    else:
        os.symlink(src, dst)

def copyFileOrTree(src, dst):
    """
    Copy src to dst, where the two can both be files or directories.
    If `src` and `dst` are directories, `dst` must not exist yet.

    :param src: (str) path to a source file or directory
    :param dst: (str) path to a destination file or directory.
    :return: none
    """
    if getParamAsBoolean('GCAM.CopyAllFiles') and src[0] == '.':   # convert relative paths
        src = unixPath(os.path.normpath(os.path.join(os.path.dirname(dst), src)))

    if os.path.islink(src):
        src = os.readlink(src)

    if os.path.isdir(src):
        removeTreeSafely(dst)
        shutil.copytree(src, dst)
    else:
        shutil.copy2(src, dst)

# used only in gcamtool modules
# TBD: rename to removeTree
def removeTreeSafely(path, ignore_errors=True):
    if not os.path.lexists(path):
        return

    refWorkspace = os.path.realpath(getParam('GCAM.RefWorkspace'))
    thisPath = os.path.realpath(path)
    if os.path.commonprefix((refWorkspace, thisPath)) == refWorkspace:
        raise PygcamException("Refusing to delete %s, which is part of the reference workspace" % path)

    _logger.debug("shutil.rmtree('%s')", thisPath)
    shutil.rmtree(thisPath, ignore_errors=ignore_errors)

def removeFileOrTree(path, raiseError=True):
    """
    Remove a file or an entire directory tree. Handles removal of symlinks
    on Windows, as these are treated differently in that system.

    :param path: (str) the pathname of a file or directory.
    :param raiseError: (bool) if True, re-raise any error that occurs
       during the file operations, else errors are ignored.
    :return: none
    """
    from .windows import removeSymlink

    if not os.path.lexists(path):
        return

    try:
        if os.path.islink(path):
            # Windows treats links to files and dirs differently.
            # NB: if not on Windows, just calls os.remove()
            removeSymlink(path)
        else:
            if os.path.isdir(path):
                removeTreeSafely(path)
            else:
                os.remove(path)
    except Exception as e:
        if raiseError:
            raise

def systemOpenFile(path):
    """
    Ask the operating system to open a file at the given pathname.

    :param path: (str) the pathname of a file to open
    :return: none
    """
    import platform
    from subprocess import call

    if platform.system() == 'Windows':
        call(['start', os.path.abspath(path)], shell=True)
    else:
        # "-g" => don't bring app to the foreground
        call(['open', '-g', path], shell=False)

def coercible(value, type, raiseError=True):
    """
    Attempt to coerce a value to `type` and raise an error on failure.

    :param value: any value coercible to `type`
    :param type: any Python type
    :return: (`type`) the coerced value, if it's coercible, otherwise
       None if raiseError is False
    :raises PygcamException: if not coercible and raiseError is True
    """
    try:
        value = type(value)
    except (TypeError, ValueError) as e:
        if raiseError:
            raise PygcamException("%s: %r is not coercible to %s" % (getFuncName(1), value, type.__name__))
        else:
            return None

    return value

def unixPath(path, rmFinalSlash=False):
    """
    Convert a path to use Unix-style slashes, optionally
    removing the final slash, if present.

    :param path: (str) a pathname
    :param rmFinalSlash: (bool) True if a final slash should
           be removed, if present.
    :return: (str) the modified pathname
    """
    if IsWindows:
        path = path.replace('\\', '/')

    if rmFinalSlash and path[-1] == '/':
        path = path[0:-1]

    return path

def pathjoin(*elements, **kwargs):
    path = os.path.join(*elements)

    if kwargs.get('normpath'):
        path = os.path.normpath(path)

    if kwargs.get('realpath'):
        path = os.path.realpath(path)

    return unixPath(path, rmFinalSlash=True)

def shellCommand(command, shell=True, raiseError=True):
    """
    Run a shell command and optionally raise PygcamException error.

    :param command: the command to run, with arguments. This can be expressed
      either as a string or as a list of strings.
    :param shell: if True, run `command` in a shell, otherwise run it directly.
    :param raiseError: if True, raise `ToolError` on command failure.
    :return: exit status of executed command
    :raises: ToolError
    """
    _logger.info(command)
    exitStatus = subprocess.call(command, shell=shell)
    if exitStatus != 0:
        if raiseError:
            raise PygcamException("\n*** Command failed: %s\n*** Command exited with status %s\n" % (command, exitStatus))

    return exitStatus

def flatten(listOfLists):
    """
    Flatten one level of nesting given a list of lists. That is, convert
    [[1, 2, 3], [4, 5, 6]] to [1, 2, 3, 4, 5, 6].

    :param listOfLists: a list of lists, obviously
    :return: the flattened list
    """
    from itertools import chain

    return list(chain.from_iterable(listOfLists))

def ensureExtension(filename, ext):
    """
    Force a filename to have the given extension, `ext`, adding it to
    any other extension, if present. That is, if `filename` is ``foo.bar``,
    and `ext` is ``baz``, the result will be ``foo.bar.baz``.
    If `ext` doesn't start with a ".", one is added.

    :param filename: filename
    :param ext: the desired filename extension
    :return: filename with extension `ext`
    """
    mainPart, extension = os.path.splitext(filename)
    ext = ext if ext[0] == '.' else '.' + ext

    if not extension:
        filename = mainPart + ext
    elif extension != ext:
        filename += ext

    return filename

def ensureCSV(file):
    """
    Ensure that the file has a '.csv' extension by replacing or adding
    the extension, as required.

    :param file: (str) a filename
    :return: (str) the filename with a '.csv' extension.
    """
    return ensureExtension(file, '.csv')

def getYearCols(years, timestep=5):
    """
    Generate a list of names of year columns in GCAM result files from a
    string indicating a year range.

    :param years: (str) a string of the form "2020-2050"
    :param timestep: (int) the number of years between timesteps
    :return: (list of strings) the names of the corresponding columns
    """
    try:
        yearRange = map(int, years.split('-'))
        if not len(yearRange) == 2:
            raise Exception
    except:
        raise Exception('Years must be specified as two years separated by a hyphen, as in "2020-2050"')

    cols = map(str, range(yearRange[0], yearRange[1]+1, timestep))
    return cols

def saveToFile(txt, dirname='', filename=''):
    """
    Save the given text to a file in the given directory.

    :param txt: (str) the text to save
    :param dirname: (str) path to a directory
    :param filename: (str) the name of the file to create

    :return: none
    """
    if dirname:
        mkdirs(dirname)

    pathname = pathjoin(dirname, filename)

    _logger.debug("Writing %s", pathname)
    with open(pathname, 'w') as f:
        f.write(txt)

QueryResultsDir = 'queryResults'

def getBatchDir(scenario, resultsDir):
    """
    Get the name of the directory holding batch query results..

    :param scenario: (str) the name of a scenario
    :param resultsDir: (str) the directory in which the batch
        results directory should be created
    :return: (str) the pathname to the batch results directory
    """
    pathname = pathjoin(resultsDir, scenario, QueryResultsDir)
    return pathname


def mkdirs(newdir, mode=0o770):
    """
    Try to create the full path `newdir` and ignore the error if it already exists.

    :param newdir: the directory to create (along with any needed parent directories)
    :return: nothing
    """
    from errno import EEXIST

    try:
        os.makedirs(newdir, mode)
    except OSError as e:
        if e.errno != EEXIST:
            raise

def getExeDir(workspace, chdir=False):
    workspace = os.path.abspath(os.path.expanduser(workspace))     # handle ~ in pathname
    exeDir    = pathjoin(workspace, 'exe')

    if chdir:
        _logger.info("cd %s", exeDir)
        os.chdir(exeDir)

    return exeDir

def loadModuleFromPath(modulePath, raiseOnError=True):
    """
    Load a module from a '.py' or '.pyc' file from a path that ends in the
    module name, i.e., from "foo/bar/Baz.py", the module name is 'Baz'.

    :param modulePath: (str) the pathname of a python module (.py or .pyc)
    :param raiseOnError: (bool) if True, raise an error if the module cannot
       be loaded
    :return: (module) a reference to the loaded module, if loaded, else None.
    :raises: PygcamException
    """
    from imp import load_source, load_compiled  # lazy import to speed startup

    # Extract the module name from the module path
    modulePath = unixPath(modulePath)
    base       = os.path.basename(modulePath)
    moduleName = base.split('.')[0]

    _logger.debug('loading module %s' % modulePath)

    # Load the compiled code if it's a '.pyc', otherwise load the source code
    module = None
    try:
        module = load_source(moduleName, modulePath)
    except Exception as e:
        errorString = "Can't load module %s from path %s: %s" % (moduleName, modulePath, e)
        if raiseOnError:
            raise PygcamException(errorString)

        _logger.error(errorString)

    return module


def importFrom(modname, objname, asTuple=False):
    """
    Import `modname` and return reference to `objname` within the module.

    :param modname: (str) the name of a Python module
    :param objname: (str) the name of an object in module `modname`
    :param asTuple: (bool) if True a tuple is returned, otherwise just the object
    :return: (object or (module, object)) depending on `asTuple`
    """
    from importlib import import_module

    module = import_module(modname, package=None)
    obj    = getattr(module, objname)
    return ((module, obj) if asTuple else obj)


def importFromDotSpec(spec):
    """
    Import an object from an arbitrary dotted sequence of packages, e.g.,
    "a.b.c.x" by splitting this into "a.b.c" and "x" and calling importFrom().

    :param spec: (str) a specification of the form package.module.object
    :return: none
    :raises PygcamException: if the import fails
    """
    modname, objname = spec.rsplit('.', 1)

    try:
        return importFrom(modname, objname)

    except ImportError:
        raise PygcamException("Can't import '%s' from '%s'" % (objname, modname))

# Deprecated?
def readScenarioName(configFile):
    """
    Read the file `configFile` and extract the scenario name.

    :param configFile: (str) the path to a GCAM configuration file
    :return: (str) the name of the scenario defined in `configFile`
    """
    parser = ET.XMLParser(remove_blank_text=True)
    tree   = ET.parse(configFile, parser)
    scenarioName = tree.find('//Strings/Value[@name="scenarioName"]')
    return scenarioName.text

def printSeries(series, label, header='', asStr=False):
    """
    Print a `series` of values, with a give `label`.

    :param series: (convertible to pandas Series) the values
    :param label: (str) a label to print for the data
    :return: none
    """
    import pandas as pd

    if type(series) == pd.DataFrame:
        df = series
        df = df.T
    else:
        df = pd.DataFrame(pd.Series(series))  # DF is more convenient for printing

    df.columns = [label]

    oldPrecision = pd.get_option('precision')
    pd.set_option('precision', 5)
    s = "%s\n%s" % (header, df.T)
    pd.set_option('precision', oldPrecision)

    if asStr:
        return s
    else:
        print(s)


TRIAL_STRING_DELIMITER = ','

# Unused
def parseTrialString(string):
    """
    Converts a comma-separated list of ranges into a list of numbers.
    Ex. 1,3,4-6,2 becomes [1,3,4,5,6,2]. Duplicates are deleted. This
    function is the inverse of :func:`createTrialString`.

    :param string: (str) comma-separate list of ints or int ranges indicated
      by two ints separated by a hyphen.
    :return: (list) a list of ints
    """
    rangeStrs = string.split(TRIAL_STRING_DELIMITER)
    res = set()
    for rangeStr in rangeStrs:
        r = [int(x) for x in rangeStr.strip().split('-')]
        if len(r) == 2:
            r = range(r[0], r[1] + 1)
        elif len(r) != 1:
            raise ValueError('Malformed trial string.')
        res = res.union(set(r))
    return list(res)

# Unused
def createTrialString(lst):
    '''
    Assemble a list of numbers into a compact list using hyphens to identify ranges.
    This function is the inverse of :func:`parseTrialString`.
    '''
    from itertools import groupby   # lazy imports
    from operator import itemgetter

    lst = sorted(set(lst))
    ranges = []
    for _, g in groupby(enumerate(lst), lambda (i, x): i - x):
        group = map(lambda x: str(itemgetter(1)(x)), g)
        if group[0] == group[-1]:
            ranges.append(group[0])
        else:
            ranges.append(group[0] + '-' + group[-1])
    return TRIAL_STRING_DELIMITER.join(ranges)

# Unused
def chunkify(lst, chunks):
    """
    Iterator to turn a list into the number of lists given by chunks.
    In the case that len(lst) % chunksize != 0, all chunks are made as
    close to the same size as possible.

    :param lst: (list) a list of values
    :param chunks: (int) the number of chunks to create
    :return: iterator that returns one chunk at a time
    """
    l = len(lst)
    numWithExtraEntry = l % chunks  # this many chunks have one extra entry
    chunkSize = (l - numWithExtraEntry) / chunks + 1
    switch = numWithExtraEntry * chunkSize

    i = 0
    while i < l:
        if i == switch:
            chunkSize -= 1
        yield lst[i:i + chunkSize]
        i += chunkSize



# Copyright (c) 2012-2016. The Regents of the University of California (Regents)
# and Richard Plevin. See the file COPYRIGHT.txt for details.
import os
import time
import ipyparallel as ipp

from pygcam.config import getConfig, getParam, setParam, getParamAsFloat, getParamAsBoolean
from pygcam.error import GcamError, GcamSolverError
from pygcam.log import getLogger, configureLogs
from pygcam.signals import (catchSignals, TimeoutSignalException, UserInterruptException)
from pygcam.utils import mkdirs

from pygcam.mcs.constants import RUNNER_SUCCESS, RUNNER_FAILURE
from pygcam.mcs.context import Context
from pygcam.mcs.error import PygcamMcsUserError, GcamToolError
from pygcam.mcs.Database import (RUN_SUCCEEDED, RUN_FAILED, RUN_KILLED, RUN_ABORTED,
                                 RUN_UNSOLVED, RUN_GCAMERROR, RUN_RUNNING)
from pygcam.mcs.util import readTrialDataFile, symlink
from pygcam.mcs.XMLParameterFile import XMLParameter, XMLParameterFile

_logger = getLogger(__name__)

def _secondsToStr(t):
    minutes, seconds = divmod(t, 60)
    hours, minutes   = divmod(minutes, 60)
    return "%d:%02d:%02d" % (hours, minutes, seconds)


def _runPygcamSteps(steps, context, runWorkspace=None, raiseError=True):
    """
    run "gt +P {project} --mcs=trial run -s {step[,step,...]} -S {scenarioName} ..."
    For Monte Carlo trials.
    """
    import pygcam.tool

    runWorkspace = runWorkspace or getParam('MCS.RunWorkspace')

    trialDir = context.getTrialDir()
    groupArg = ['-g', context.groupName] if context.groupName else []

    # N.B. MCS.RunWorkspace is the RefWorkspace for trial sandboxes
    toolArgs = ['+P', context.projectName, '--mcs=trial',
                '--set=GCAM.SandboxRefWorkspace=' + runWorkspace,
                'run', '-s', steps, '-S', context.scenario,
                '--sandboxDir=' + trialDir] + groupArg

    command = 'gt ' + ' '.join(toolArgs)
    _logger.debug('Running: %s', command)
    status = pygcam.tool.main(argv=toolArgs, raiseError=True)
    msg = '"%s" exited with status %d' % (command, status)

    if status != 0 and raiseError:
        raise GcamToolError(msg)

    _logger.info("_runSteps: " + msg)
    return status

def _readParameterInfo(context, paramPath):
    from pygcam.xmlSetup import ScenarioSetup

    scenarioFile  = getParam('GCAM.ScenarioSetupFile')
    scenarioSetup = ScenarioSetup.parse(scenarioFile)
    scenarioNames = scenarioSetup.scenariosInGroup(context.groupName)

    paramFile = XMLParameterFile(paramPath)
    paramFile.loadInputFiles(context, scenarioNames, writeConfigFiles=False)
    paramFile.runQueries()
    return paramFile

def _applySingleTrialData(df, context, paramFile):
    simId    = context.simId
    trialNum = context.trialNum
    trialDir = context.getTrialDir(create=True)

    _logger.info('_applySingleTrialData for %s, %s', context, paramFile.filename)
    XMLParameter.applyTrial(simId, trialNum, df)   # Update all parameters as required
    paramFile.writeLocalXmlFiles(trialDir)         # N.B. creates trial-xml subdir

    linkDest = os.path.join(trialDir, 'local-xml')
    _logger.info('creating symlink to %s', linkDest)
    symlink('../../../../Workspace/local-xml', linkDest)


def _runGcamTool(context, noGCAM=False, noBatchQueries=False,
                noPostProcessor=False):
    '''
    Run GCAM in the current working directory and return exit status.
    '''
    from XMLParameterFile import XMLParameter, decache

    _logger.debug("_runGcamTool: %s", context)

    # For running in an ipyparallel engine, forget instances from last run
    decache()

    # TBD: #### set to True to help debug ipyparallel issues ####
    debuggingOnly = False
    if debuggingOnly:
        time.sleep(30)
        return RUNNER_SUCCESS

    simId = context.simId
    baselineName = context.baseline
    isBaseline = not baselineName

    if isBaseline and not noGCAM:
        paramPath = getParam('MCS.ParametersFile')      # TBD: gensim has optional override of param file. Keep it?
        paramFile = _readParameterInfo(context, paramPath)

        df = readTrialDataFile(simId)
        columns = df.columns

        # add data for linked columns if not present
        linkPairs = XMLParameter.getParameterLinks()
        for linkName, dataCol in linkPairs:
            if linkName not in columns:
                df[linkName] = df[dataCol]

        _applySingleTrialData(df, context, paramFile)

    if noGCAM:
        _logger.info('_runGcamTool: skipping GCAM')
        gcamStatus = 0
    else:
        start = time.time()

        # N.B. setup step calls pygcam.setup.setupWorkspace
        gcamStatus = _runPygcamSteps('setup,prequery,gcam', context)

        stop = time.time()
        _logger.info("_runGcamTool: elapsed time: %s", _secondsToStr(stop - start))

    if gcamStatus == 0:
        if not noBatchQueries:
            _runPygcamSteps('query', context)

        if not noPostProcessor:
            steps = getParam('MCS.PostProcessorSteps')     # e.g., "diff,CI"
            if steps:
                _runPygcamSteps(steps, context)

        status = RUNNER_SUCCESS
    else:
        status = RUNNER_FAILURE

    _logger.info("_runGcamTool: exiting with status %d", status)
    return status


class WorkerResult(object):
    '''
    Encapsulates the results returned from a worker task.
    '''
    def __init__(self, context, errorMsg):
        from .XMLResultFile import collectResults, RESULT_TYPE_SCENARIO, RESULT_TYPE_DIFF

        self.context  = context
        self.errorMsg = errorMsg
        self.resultsList = []

        if context.status == RUN_SUCCEEDED:
            self.resultsList = collectResults(context, RESULT_TYPE_SCENARIO)

            if context.baseline:  # also save 'diff' results
                diffResults = collectResults(context, RESULT_TYPE_DIFF)
                if diffResults:
                    self.resultsList += diffResults

            _logger.debug('Worker results saving %s', self.resultsList)


    def __str__(self):
        c = self.context
        return "<WorkerResult run=%s sim=%s trial=%s, scenario=%s, status=%s error=%s>" % \
               (c.runId, c.simId, c.trialNum, c.scenario, c.status, self.errorMsg)


class Worker(object):
    '''
    Defines the methods and data associated with a worker task.
    '''
    def __init__(self, context, argDict):
        """
        Initialize a Worker instance

        :param context: (Context) description of trial to run
        :param argDict: (dict) various args passed from command-line
        """
        getConfig()
        configureLogs()

        catchSignals()
        # signal.signal(signal.SIGUSR1, _handleSIGUSR1)

        self.errorMsg = None
        self.context  = context
        self.argDict  = argDict
        self.runLocal = argDict.get('runLocal', False)

    def runTrial(self):
        """
        Run a single trial on the current engine using the local Worker.

        :return: (WorkerResult) holds run identification info and completion status
        """
        context = self.context
        runDir = context.getScenarioDir(create=True)
        _logger.info("runDir is %s", runDir)
        os.chdir(runDir)

        trialDir = os.path.dirname(runDir)
        logDir = os.path.join(trialDir, 'log')
        mkdirs(logDir)

        if not self.runLocal:
            logFile = os.path.join(logDir, context.scenario + '.log')
            setParam('GCAM.LogFile', logFile)
            setParam('GCAM.LogConsole', 'False')    # avoids duplicate output to file
            configureLogs(force=True)

            self.setStatus(RUN_RUNNING)

        result = self._runTrial()
        return result

    def setStatus(self, status):
        from ipyparallel.datapub import publish_data

        context = self.context
        context.setVars(status=status)

        if not self.runLocal:
            publish_data(dict(context=context))

    def _runTrial(self):
        """
        Run a single Monte Carlo trial.

        :return: (WorkerResult) Contains execution status, one of {'succeeded', 'failed', 'alarmed', 'aborted', 'killed'},
           as well as Context, any error message, and a list of results to post to the database.
        """
        context = self.context
        argDict = self.argDict

        noGCAM          = argDict.get('noGCAM', False)
        noBatchQueries  = argDict.get('noBatchQueries', False)
        noPostProcessor = argDict.get('noPostProcessor', False)

        trialNum = context.trialNum
        errorMsg = None

        _logger.info('Running trial %d' % trialNum)
        try:
            exitCode = _runGcamTool(context, noGCAM=noGCAM,
                                    noBatchQueries=noBatchQueries,
                                    noPostProcessor=noPostProcessor)
            status = RUN_SUCCEEDED if exitCode == 0 else RUN_FAILED

        except TimeoutSignalException:
            errorMsg = "Trial %d terminated by system" % trialNum
            status = RUN_KILLED

        # except AlarmSignalException:
        #     errorMsg = "Trial %d terminated by internal alarm" % trialNum
        #     status = RUN_ALARMED

        except UserInterruptException:
            errorMsg = "Interrupted by user"
            status = RUN_KILLED

        except GcamToolError as e:
            errorMsg = "%s" % e
            status = RUN_FAILED

        except PygcamMcsUserError as e:
            errorMsg = "%s" % e
            status = RUN_FAILED

        except GcamSolverError as e:
            errorMsg = "%s" % e
            status = RUN_UNSOLVED

        except GcamError as e:
            errorMsg = "%s" % e
            status = RUN_GCAMERROR

        except Exception as e:
            errorMsg = "%s" % e
            status = RUN_ABORTED     # run-time error in loaded module

        if errorMsg:
            _logger.error("Trial status: %s: %s", status, errorMsg)
            if status == RUN_ABORTED and getParamAsBoolean('GCAM.ShowStackTrace'):
                import traceback
                errorMsg = traceback.format_exc()
                _logger.debug(errorMsg)
        else:
            _logger.info('Trial status: %s', status)

        self.setStatus(status)
        result = WorkerResult(context, errorMsg)
        return result


latestStartTime = None

def runTrial(context, argDict):
    '''
    Remotely-callable function providing an interface to the Worker
    class.

    :param context: (Context) information describing the run
    :param argDict: (dict) with bool values for keys 'runLocal',
        'noGCAM', 'noBatchQueries', and 'noPostProcessor'
    :return: (WorkerResult) run identification info and completion status
    '''
    global latestStartTime

    if not argDict.get('runLocal', False):
        # On the first run, compute the latest time we should start a new trial.
        # On subsequent runs, check that there's adequate time still left.
        if latestStartTime is None:
            startTime = time.time()

            wallTime  = os.getenv('MCS_WALLTIME', '2:00') # should always be set except when debugging
            parts = map(int, wallTime.split(':'))
            secs = parts.pop()
            mins = parts.pop() if parts else 0
            hrs  = parts.pop() if parts else 0

            minTimeToRun = getParamAsFloat('IPP.MinTimeToRun')
            latestStartTime = (startTime + secs + 60 * mins + 3600 * hrs) - (minTimeToRun * 60)

        else:
            if time.time() > latestStartTime:
                # TBD: test this!
                # raising UnmetDependency error causes scheduler to reassign to another engine
                raise ipp.UnmetDependency()

                # context.setVars(status=ENG_TERMINATE) # tell master to terminate us
                # time.sleep(10) # don't consume queue while waiting for termination
                # return WorkerResult(context, 'insufficient time remaining')

    worker = Worker(context, argDict)
    result = worker.runTrial()
    return result


if __name__ == '__main__':
    context = Context(runId=1001, simId=1, trialNum=2, scenario='baseline',
                      projectName='paper1', groupName='mcs', store=False)

    argDict = {'runLocal': True,
               'noGCAM': False,
               'noBatchQueries': False,
               'noPostProcessor': False}
    result = runTrial(context, argDict)
    print(result)



'''
Created on Dec 20, 2012

@author: Rich Plevin
@author: Sam Fendell

Copyright (c) 2012-2015. The Regents of the University of California (Regents)
and Richard Plevin. See the file COPYRIGHT.txt for details.

Implements the Latin Hypercube Sampling technique as described by Iman and Conover, 1982,
including correlation control both for no correlation or for a specified correlation
matrix for the sampled parameters.

Heavily modified from http://nullege.com/codes/show/src@m@o@model-builder-HEAD@Bayes@lhs.py
'''
import numpy as np
from scipy import stats
from pandas import DataFrame

def rankCorrCoef(m):
    '''
    Take a 2-D array of values and produce a array of rank correlation
    coefficients representing the rank correlation among the columns.
    '''
    dummy, cols = m.shape
    corrCoef = np.zeros((cols, cols))  # @UndefinedVariable

    for i in range(cols):
        corrCoef[i, i] = 1.  # All columns are perfectly correlated with themselves
        for j in range(i + 1, cols):
            corr = stats.spearmanr(m[:, i], m[:, j])[0]
            corrCoef[i, j] = corrCoef[j, i] = corr

    return corrCoef


def genRankValues(params, trials, corrMat):
    '''
    Generate a data set of 'trials' ranks for 'params'
    parameters that obey the given correlation matrix.

    params: integer denoting number of parameters.

    trials: integer denoting number of trials.

    corrMat: rank correlation matrix for parameters.
    corrMat[i,j] denotes the rank correlation between parameter
    i and j.

    Output is a matrix with 'trials' rows and 'params' columns.
    The i'th column represents the ranks for the i'th parameter.

    So an input with params=3 and trials=6 might output:

    [[1,4,6],
     [2,3,5],
     [4,1,3],
     [6,5,2],
     [5,2,1],
     [3,6,4]]
    '''
    # Create van der Waarden scores
    strata = np.arange(1.0, trials + 1) / (trials + 1)
    vdwScores = stats.norm().ppf(strata)

    S = np.zeros((trials, params))
    for i in xrange(params):
        np.random.shuffle(vdwScores)
        S[:, i] = vdwScores

    P = np.linalg.cholesky(corrMat)

    E = rankCorrCoef(S)
    Q = np.array(np.linalg.cholesky(E))
    final = np.dot(np.dot(S, np.linalg.inv(Q).T), P.T)

    ranks = np.zeros((trials, params), dtype='i')
    for i in xrange(params):
        ranks[:, i] = stats.rankdata(final[:, i])

    return ranks


def getPercentiles(trials=100):
    '''
    Generate a list of 'trials' values, one from each of 'trials' equal-size
    segments from a uniform distribution. These are used with an RV's ppf
    (percent point function = inverse cumulative function) to retrieve the
    values for that RV at the corresponding percentiles.
    '''
    segmentSize = float(1. / trials)
    points = stats.uniform.rvs(size=trials) * segmentSize + np.arange(trials) * segmentSize  # @UndefinedVariable
    return points


def lhs(paramList, trials, corrMat=None, columns=None, skip=None):
    """
    Produce an ndarray or DataFrame of 'trials' rows of values for the given parameter
    list, respecting the correlation matrix 'corrMat' if one is specified, using Latin
    Hypercube (stratified) sampling.

    The values in the i'th column are drawn from the ppf function of the i'th parameter
    from paramList, and each columns i and j are rank correlated according to corrMat[i,j].

    :param paramList: (list of rv-like objects representing parameters) Only requirement
           on parameter objects is that they must implement the ppf function.
    :param trials: (int) number of trials to generate for each parameter.
    :param corrMat: a numpy matrix representing the correlation between the parameters.
           corrMat[i,j] should give the correlation between the i'th and j'th
           entries of paramlist.
    :param columns: (None or list(str)) Column names to use to return a DataFrame.
    :param skip: (list of params)) Parameters to process later because they are
           dependent on other parameter values (e.g., they're "linked"). These
           cannot be correlated.
    :return: ndarray or DataFrame with `trials` rows of values for the `paramList`.
    """
    ranks = genRankValues(len(paramList), trials, corrMat) if corrMat is not None else None

    samples = np.zeros((trials, len(paramList)))  # @UndefinedVariable

    skip = skip or []

    for i, param in enumerate(paramList):
        if param in skip:
            continue    # process later

        values = param.ppf(getPercentiles(trials))  # extract values from the RV for these percentiles

        if corrMat is None:
            np.random.shuffle(values)  # randomize the stratified samples
        else:
            indices = ranks[:, i] - 1  # make them 0-relative
            values = values[indices]   # reorder to respect correlations

        samples[:, i] = values

    return DataFrame(samples, columns=columns) if columns else samples

def lhsAmend(df, rvList, trials):
    """
    Amend the DataFrame with LHS data by adding columns for the given parameters.
    This allows "linked" parameters to refer to values of other parameters.

    :param df: (DataFrame) Generated by prior call to LHS or something similar.
    :param paramList: (list of params) The parameters to fill in the df with
    :param trials: (int) the number of trials to generate for each parameter
    :return: none
    """
    for rv in rvList:
        values = rv.ppf(getPercentiles(trials))  # extract values from the RV for these percentiles
        if not isinstance(values, np.ndarray):
            values = values.as_matrix()         # convert pandas Series if needed
        np.random.shuffle(values)                   # randomize the stratified samples
        param = rv.getParameter()
        paramName = param.getName()
        df[paramName] = values
        #continue



'''
  .. Copyright (c) 2016 Richard Plevin

  See the https://opensource.org/licenses/MIT for license details.
'''
import os
from .log import getLogger
from .error import CommandlineError, FileFormatError
from .utils import mkdirs, pathjoin, ensureCSV, QueryResultsDir
from .query import readCsv, dropExtraCols, csv2xlsx, sumYears, sumYearsByGroup, QueryFile

_logger = getLogger(__name__)


def computeDifference(df1, df2, resetIndex=True, dropna=True):
    """
    Compute the difference between two DataFrames.

    :param df1: a pandas DataFrame instance
    :param obj2: a pandas DataFrame instance
    :param resetIndex: (bool) if True (the default), the index in the DataFrame
      holding the computed difference is reset so that data in non-year columns
      appear in individual columns. Otherwise, the index in the returned
      DataFrame is based on all non-year columns.
    :param dropna: (bool) if True, drop rows with NaN values after computing difference
    :return: a pandas DataFrame with the difference in all the year columns, computed
      as (df2 - df1).
    """
    df1 = dropExtraCols(df1, inplace=False)
    df2 = dropExtraCols(df2, inplace=False)

    if set(df1.columns) != set(df2.columns):
        raise FileFormatError("Can't compute difference because result sets have different columns. df1:%s, df2:%s" \
                              % (df1.columns, df2.columns))

    # Handle corner case in which query results for non-existent data have zero in Units column
    if 'Units' in df1.columns:
        units = list(df1.Units.unique())
        if len(units) == 2 and '0.0' in units:
            units.remove('0.0')
            realUnits = units[0]
            df1.Units = realUnits
            df2.Units = realUnits

    yearCols = filter(str.isdigit, df1.columns)
    nonYearCols = list(set(df1.columns) - set(yearCols))

    df1.set_index(nonYearCols, inplace=True)
    df2.set_index(nonYearCols, inplace=True)

    # Compute difference for timeseries values
    diff = df2 - df1

    if dropna:
        diff.dropna(inplace=True)

    if resetIndex:
        diff.reset_index(inplace=True)      # convert multi-index back to regular column values

    return diff


def writeDiffsToCSV(outFile, referenceFile, otherFiles, skiprows=1, interpolate=False,
                    years=None, startYear=0):
    """
    Compute the differences between the data in a reference .CSV file and one or more other
    .CSV files as (other - reference), optionally interpolating annual values between
    timesteps, storing the results in a single .CSV file.
    See also :py:func:`writeDiffsToXLSX` and :py:func:`writeDiffsToFile`

    :param outFile: (str) the name of the .CSV file to create
    :param referenceFile: (str) the name of a .CSV file containing reference results
    :param otherFiles: (list of str) the names of other .CSV file for which to
       compute differences.
    :param skiprows: (int) should be 1 for GCAM files, to skip header info before column names
    :param interpolate: (bool) if True, linearly interpolate annual values between timesteps
       in all data files and compute the differences for all resulting years.
    :param years: (iterable of 2 values coercible to int) the range of years to include in
       results.
    :param startYear: (int) the year at which to begin interpolation, if interpolate is True.
       Defaults to the first year in `years`.
    :return: none
    """
    refDF = readCsv(referenceFile, skiprows=skiprows, interpolate=interpolate,
                    years=years, startYear=startYear)

    with open(outFile, 'w') as f:
        for otherFile in otherFiles:
            otherFile = ensureCSV(otherFile)   # add csv extension if needed
            otherDF   = readCsv(otherFile, skiprows=skiprows, interpolate=interpolate,
                                years=years, startYear=startYear)

            diff = computeDifference(refDF, otherDF)
            csvText = diff.to_csv(index=None)
            label = "[%s] minus [%s]" % (otherFile, referenceFile)
            f.write("%s\n%s" % (label, csvText))    # csvText has "\n" already


def writeDiffsToXLSX(outFile, referenceFile, otherFiles, skiprows=1, interpolate=False,
                     years=None, startYear=0):
    """
    Compute the differences between the data in a reference .CSV file and one or more other
    .CSV files as (other - reference), optionally interpolating annual values between
    timesteps, storing the results in a single .XLSX file with each difference matrix
    on a separate worksheet, and with an index worksheet with links to the other worksheets.
    See also :py:func:`writeDiffsToCSV` and :py:func:`writeDiffsToFile`.

    :param outFile: (str) the name of the .XLSX file to create
    :param referenceFile: (str) the name of a .CSV file containing reference results
    :param otherFiles: (list of str) the names of other .CSV file for which to
       compute differences.
    :param skiprows: (int) should be 1 for GCAM files, to skip header info before column names
    :param interpolate: (bool) if True, linearly interpolate annual values between timesteps
       in all data files and compute the differences for all resulting years.
    :param years: (iterable of 2 values coercible to int) the range of years to include in
       results.
    :param startYear: (int) the year at which to begin interpolation, if interpolate is True.
       Defaults to the first year in `years`.
    :return: none
    """
    import pandas as pd

    with pd.ExcelWriter(outFile, engine='xlsxwriter') as writer:
        sheetNum = 1
        _logger.debug("Reading reference file:", referenceFile)
        refDF = readCsv(referenceFile, skiprows=skiprows, interpolate=interpolate,
                        years=years, startYear=startYear)

        for otherFile in otherFiles:
            otherFile = ensureCSV(otherFile)   # add csv extension if needed
            _logger.debug("Reading other file:", otherFile)
            otherDF = readCsv(otherFile, skiprows=skiprows, interpolate=interpolate,
                              years=years, startYear=startYear)

            sheetName = 'Diff%d' % sheetNum
            sheetNum += 1

            diff = computeDifference(refDF, otherDF)
            diff.to_excel(writer, index=None, sheet_name=sheetName, startrow=2, startcol=0)

            worksheet = writer.sheets[sheetName]
            label     = "[%s] minus [%s]" % (otherFile, referenceFile)
            worksheet.write_string(0, 0, label)

            startRow = diff.shape[0] + 4
            worksheet.write_string(startRow, 0, otherFile)
            startRow += 2
            otherDF.reset_index(inplace=True)
            otherDF.to_excel(writer, index=None, sheet_name=sheetName, startrow=startRow, startcol=0)

        dropExtraCols(refDF, inplace=True)
        _logger.debug("writing DF to excel file", outFile)
        refDF.to_excel(writer, index=None, sheet_name='Reference', startrow=0, startcol=0)


def writeDiffsToFile(outFile, referenceFile, otherFiles, ext='csv', skiprows=1, interpolate=False,
                     years=None, startYear=0):
    """
    Compute the differences between the data in a reference .CSV file and one or more other
    .CSV files as (other - reference), optionally interpolating annual values between
    timesteps, storing the results in a single .CSV or .XLSX file. See :py:func:`writeDiffsToCSV`
    and :py:func:`writeDiffsToXLSX` for more details.

    :param outFile: (str) the name of the file to create
    :param referenceFile: (str) the name of a .CSV file containing reference results
    :param otherFiles: (list of str) the names of other .CSV file for which to
       compute differences.
    :param ext: (str) if '.csv', results are written to a single .CSV file, otherwise, they
       are written to an .XLSX file.
    :param skiprows: (int) should be 1 for GCAM files, to skip header info before column names
    :param interpolate: (bool) if True, linearly interpolate annual values between timesteps
       in all data files and compute the differences for all resulting years.
    :param years: (iterable of 2 values coercible to int) the range of years to include in
       results.
    :param startYear: (int) the year at which to begin interpolation, if interpolate is True.
       Defaults to the first year in `years`.
    :return: none
    """
    writer = writeDiffsToCSV if ext == '.csv' else writeDiffsToXLSX
    writer(outFile, referenceFile, otherFiles, skiprows=skiprows, interpolate=interpolate,
           years=years, startYear=startYear)


def diffCsvPathname(query, baseline, policy, diffsDir=None, workingDir='.', createDir=False):
    """
    Compute the path to the CSV file containing differences between `policy` and
    `baseline` scenarios for `query`.

    :param query: (str) the base file name of the query result
    :param baseline: (str) the baseline scenario
    :param policy: (str) the policy scenario
    :param workingDir: (str) the directory immediately above the baseline
        and policy sandboxes.
    :param createDir: (bool) whether to create the diffs directory, if needed.
    :return: (str) the pathname of the CSV file
    """
    diffsDir = diffsDir or pathjoin(workingDir, policy, 'diffs')
    if createDir:
        mkdirs(diffsDir)

    pathname = pathjoin(diffsDir, '%s-%s-%s.csv' % (query, policy, baseline))
    return pathname

def queryCsvPathname(query, scenario, workingDir='.'):
    """
    Compute the path to the CSV file containing results for the given
    `query` and `scenario`.

    :param query: (str) the base file name of the query result
    :param scenario: (str) the scenario name
    :param workingDir: (str) the directory immediately above the baseline
        and policy sandboxes.
    :return: (str) the pathname of the CSV file
    """
    pathname = pathjoin(workingDir, scenario, QueryResultsDir, '%s-%s.csv' % (query, scenario))
    return pathname

def diffMain(args):
    workingDir = args.workingDir
    mkdirs(workingDir)
    os.chdir(workingDir)

    _logger.debug('Working dir: %s', workingDir)

    convertOnly = args.convertOnly
    skiprows    = args.skiprows
    interpolate = args.interpolate
    groupSum    = args.groupSum
    sum         = args.sum
    queryFile   = args.queryFile
    yearStrs    = args.years.split('-')

    if len(yearStrs) == 2:
        years = yearStrs
        startYear = args.startYear

    # If a query file is given, we loop over the query names, computing required arguments to performDiff().
    if queryFile:
        if len(args.csvFiles) != 2:
            raise CommandlineError("When --queryFile is specified, 2 positional arguments--the baseline and policy names--are required.")

        baseline, policy = args.csvFiles

        # def makePath(query, scenario):
        #     return pathjoin(scenario, QueryResultsDir, '%s-%s.csv' % (query, scenario))

        mainPart, extension = os.path.splitext(queryFile)

        if extension.lower() == '.xml':
            queryFileObj = QueryFile.parse(queryFile)
            queries = queryFileObj.queryFilenames()
        else:
            with open(queryFile, 'rU') as f:    # 'U' converts line separators to '\n' on Windows
                lines = f.read()
                queries = filter(None, lines.split('\n'))   # eliminates blank lines

        for query in queries:
            baselineFile = queryCsvPathname(query, baseline, workingDir=workingDir)
            policyFile   = queryCsvPathname(query, policy,   workingDir=workingDir)

            outFile = diffCsvPathname(query, baseline, policy, workingDir=workingDir, createDir=True)
            _logger.info("Writing %s", outFile)

            writeDiffsToFile(outFile, baselineFile, [policyFile], ext='.csv', skiprows=skiprows,
                             interpolate=interpolate, years=years, startYear=startYear)
    else:
        csvFiles = map(ensureCSV, args.csvFiles)
        referenceFile = csvFiles[0]
        otherFiles    = csvFiles[1:] if len(csvFiles) > 1 else []

        outFile = args.outFile
        root, ext = os.path.splitext(outFile)
        if not ext:
            outFile = ensureCSV(outFile)
            ext = '.csv'

        extensions = ('.csv', '.xlsx')
        if ext not in extensions:
            raise CommandlineError("Output file extension must be one of %s", extensions)

        if convertOnly or groupSum or sum:
            if convertOnly:
                csv2xlsx(csvFiles, outFile, skiprows=skiprows, interpolate=interpolate)
            elif groupSum:
                sumYearsByGroup(groupSum, csvFiles, skiprows=skiprows, interpolate=interpolate)
            elif sum:
                sumYears(csvFiles, skiprows=skiprows, interpolate=interpolate)
            return

        writeDiffsToFile(outFile, referenceFile, otherFiles, ext=ext, skiprows=skiprows,
                         interpolate=interpolate, years=years, startYear=startYear)




# Copyright (c) 2016  Richard Plevin
# See the https://opensource.org/licenses/MIT for license details.

from __future__ import print_function
from six import iteritems

from pygcam.log import getLogger
from .McsSubcommandABC import McsSubcommandABC

_logger = getLogger(__name__)

def driver(args, tool):
    from pygcam.project import Project
    from ..master import Master, pidFileExists, startCluster

    if args.numTrials:
        # If the pid file doesn't exist, we assume the cluster is
        # not running and we run it with the given profile and
        # cluster ID, relying on the config file for other parameters.
        # To specify other params, use "gt cluster start" instead.
        if pidFileExists(args.profile, args.clusterId):
            _logger.warning('ipyparallel cluster is (probably) already running')
        else:
            _logger.info('Starting ipyparallel cluster')
            kwargs = {key : getattr(args, key, None) \
                      for key in ('profile', 'clusterId', 'numTrials',
                                  'maxEngines', 'minutesPerRun', 'queue')}
            startCluster(**kwargs)

    args.groupName = args.groupName or Project.defaultGroupName()

    master = Master(args)
    master.mainloop()


class RunSimCommand(McsSubcommandABC):
    def __init__(self, subparsers):
        kwargs = {'help' : '''Run the identified trials on compute engines.'''}
        super(RunSimCommand, self).__init__('runsim', subparsers, kwargs)

    def addArgs(self, parser):
        from pygcam.config import getParam, getParamAsInt, getParamAsFloat
        from pygcam.utils import ParseCommaList

        defaultProfile    = getParam('IPP.Profile')
        defaultClusterId  = getParam('IPP.ClusterId')
        defaultQueue      = getParam('IPP.Queue')
        defaultMaxEngines = getParamAsInt('IPP.MaxEngines')
        defaultMinutes    = getParamAsFloat('IPP.MinutesPerRun')
        defaultWaitSecs   = getParamAsFloat('IPP.ResultLoopWaitSecs')

        # TBD: document this variable
        defaultScenario = getParam('MCS.DefaultScenario', raiseError=False)
        scenarioHelp = 'Default value is "%s".' % defaultScenario \
                            if defaultScenario else "No default has been set."

        parser.add_argument('-a', '--addTrials', action='store_true',
                            help='''Add trials to a running cluster; don't wait for results.''')

        parser.add_argument('-B', '--noBatchQueries', action='store_true',
                            help='Skip running batch queries.')

        parser.add_argument('-c', '--clusterId', type=str, default=defaultClusterId,
                            help='''A string to identify this cluster. Default is the
                            value of config var IPP.ClusterId, currently
                            "%s".''' % defaultClusterId)

        parser.add_argument('-C', '--collectResults', action='store_true',
                            help='''Equivalent to specifying --noGCAM --noBatchQueries 
                            --noPostProcessor --runLocal. Useful if runs have actually
                            succeeded but results have not been saved to the SQL database.''')

        parser.add_argument('-D', '--noDatabase', dest='updateDatabase', action='store_false',
                            help='''Don't save query results to the SQL database.''')

        parser.add_argument('-e', '--maxEngines', type=int, default=defaultMaxEngines,
                            help='''Set maximum number of engines to create. (Ignored 
                            unless -C flag is specified.
                            Overrides config parameter IPP.MaxEngines, currently
                            %s''' % defaultMaxEngines)

        parser.add_argument('-g', '--groupName', default='',
                            help='''The name of a scenario group to process.''')

        parser.add_argument('-G', '--noGCAM', action="store_true",
                            help="Don't run GCAM, just run the batch queries and "
                                 "post-processor (if defined).")

        parser.add_argument('-i', '--shutdownWhenIdle', action='store_true',
                            help='''Shutdown engines when they are idle and there are 
                            no outstanding tasks, and shutdown controller when there are no
                            engines running.''')

        parser.add_argument('-l', '--runLocal', action='store_true',
                            help='''Runs the program locally instead of submitting a batch job.''')

        parser.add_argument('-L', '--loopOnly', action='store_true',
                            help='''Don't run any new trials; just enter the wait loop to 
                            process results.''')

        parser.add_argument('-m', '--minutesPerRun', type=int, default=defaultMinutes,
                            help='''Set the number of minutes of walltime to allocate
                            per GCAM run. Ignored unless -C flag is specified. Overrides 
                            config parameter IPP.MinutesPerRun, currently %s.''' % defaultMinutes)

        parser.add_argument('-n', '--numTrials', type=int, default=0,
                            help='''The total number of GCAM trials to be run on this cluster.''')

        parser.add_argument('-N', '--noPostProcessor', action='store_true', default=False,
                            help='''Don't run post-processor steps.''')

        parser.add_argument('-p', '--profile', type=str, default=defaultProfile,
                            help='''The name of the ipython profile to use. Default is
                            the value of config var IPP.Profile, currently
                            "%s".''' % defaultProfile)

        # If alternative flags are added, the hack in run() must be updated
        parser.add_argument('--programArgs', type=str, default=getParam('MCS.ProgramArgs'),
                            help='''Arguments to pass to user program. Quote sequences that include 
                            spaces, e.g., to pass args: -x foo, use --programArgs="-x foo"''')

        parser.add_argument('-q', '--queue', type=str, default=defaultQueue,
                            help='''The queue or partition on which to create the controller
                            and engines. Ignored unless -C flag is used. 
                            Overrides config var IPP.Queue, currently
                            "%s".''' % defaultQueue)

        parser.add_argument('-r', '--redo', dest='statuses', type=str, action=ParseCommaList,
                            help='''Re-launch all trials for the given simId with the status 
                            specified. Argument can be comma-delimited list of status names. When 
                            used with -R, trial numbers are listed but trials are not run.
                            Recognized values are {new, queued, running, failed, killed, aborted, 
                            alarmed, gcamerror, unsolved, missing}.
                            "Missing" is a pseudo-value interpreted to find all runs that have
                            not been executed, i.e., runs not appearing in the 'run' table.''')

        parser.add_argument('-R', '--redoListOnly', action='store_true', default=False,
                            help='Used with -r to only list the trials to redo, then quit.')

        parser.add_argument('-s', '--simId', type=int, default=1,
                            help='The id of the simulation (Defaults to 1.)')

        parser.add_argument('-S', '--scenario', dest='scenarios', type=str, action=ParseCommaList,
                            # required if no default is set; otherwise use default if not specified
                            required=(not defaultScenario), default=defaultScenario,
                            help='''The name of the scenario(s). May be a comma-separated list of 
                            names. Use config var MCS.DefaultScenario to set a default scenario 
                            name. ''' + scenarioHelp)

        parser.add_argument('-t', '--trials', type=str, default='',
                             help='''Comma-separated list of trial numbers and/or hyphen-separated 
                             ranges of trial numbers to run. Ex: 1,4,6-10,3. Default is to run all 
                             defined trials.''')

        parser.add_argument('-w', '--waitSecs', type=int, default=defaultWaitSecs,
                            help='''How many seconds to wait between queries to the ipyparallel
                            controller for completed jobs. Default is %d.''' % defaultWaitSecs)

        return parser   # for auto-doc generation


    def run(self, args, tool):
        if args.collectResults:
            args.noGCAM = args.noBatchQueries = args.noPostProcessor = args.runLocal = True

        if args.statuses:
            from ..Database import RUN_STATUSES
            from pygcam.error import CommandlineError

            statusSet = set(args.statuses)
            known = set(RUN_STATUSES)
            known.add('missing') # not stored in the DB, but a pseudo-status that can be specified

            unknown = statusSet - known
            if unknown:
                raise CommandlineError("Unknown status code(s): %s" % ', '.join(map(repr, unknown)))

        driver(args, tool)

#
# Test stuff
#
if __name__ == '__main__':
    import ipyparallel as ipp

    clusterId = 'mcs'
    profile = 'pygcam'

    client = ipp.Client(profile=profile, cluster_id=clusterId)
    dview = client[:]
    dview.execute('import time')

    def f(id, secs):
        from ipyparallel.engine.datapub import publish_data
        import time

        publish_data({id : 'running'})
        time.sleep(secs)
        publish_data({id : 'finished'})
        return secs

    def runStatus(res, statuses=None):
        statuses = statuses or {}

        for r in res:
            dataDict = r.data[0]
            if dataDict:
                for id, status in iteritems(dataDict):
                    statuses[id] = status
        return statuses

    def running(res, statuses=None):
        statuses = statuses or {}
        statuses = runStatus(res, statuses)
        return [id for id, status in iteritems(statuses) if status == 'running']

    dview['f'] = f

    lbview = client.load_balanced_view()
    res = [lbview.map_async(f, [i], [20]) for i in range(10)]
    m = res[0].metadata
    # md = res.metadata
    # m  = md[0]
    # kids = res._children
    # kid  = kids[0]
    status = dview.queue_status()
    print(status)




from datetime import datetime
from sqlalchemy import (Column, Integer, String, Float, Boolean,
                        ForeignKey, DateTime, UniqueConstraint, Index)
from sqlalchemy.ext.declarative import declared_attr, declarative_base
from pygcam.log import getLogger

_logger = getLogger(__name__)

#
# Object-relational mapping; defines database tables
#
ORMBase = declarative_base()

class CoreMCSMixin(object):

    @declared_attr
    def __tablename__(cls):
        return cls.__name__.lower()

# Deprecated
# class Application(CoreMCSMixin, ORMBase):
#     '''
#     Application table defining the "application" name
#     '''
#     appId   = Column(Integer, primary_key=True)
#     appName = Column(String)
#
# class Attribute(CoreMCSMixin, ORMBase):
#     '''
#     Attribute table for generic attribute-value storage.
#     '''
#     attrId      = Column(Integer, primary_key=True)
#     attrName    = Column(String)
#     attrType    = Column(String)    # should be "number" or "text" (not used currently)
#     description = Column(String, nullable=True)
#
# class AttrVal(CoreMCSMixin, ORMBase):
#     '''
#     AttrVal table for storing the values of attributes.
#     '''
#     attrId   = Column(Integer, primary_key=True)
#     runId    = Column(Integer, ForeignKey('run.runId', ondelete="CASCADE"), primary_key=True)
#     numValue = Column(Float)
#     strValue = Column(String)


class Code(CoreMCSMixin, ORMBase):
    #codeId      = Column(Integer, primary_key=True, autoincrement=False)
    codeName    = Column(String(15), primary_key=True)
    description = Column(String(256))

# TBD: add parent here so it doesn't need to be kludged in later
class Experiment(CoreMCSMixin, ORMBase):
    expId       = Column(Integer, primary_key=True)
    expName     = Column(String(30))
    description = Column(String(256))
    __table_args__ = (Index("experiment_index1", "expName", unique=True),)


class Input(CoreMCSMixin, ORMBase):
    __table_args__ = (UniqueConstraint('programId', 'paramName'),)

    inputId     = Column(Integer, primary_key=True)
    programId   = Column(Integer, ForeignKey('program.programId', ondelete="CASCADE"))
    paramName   = Column(String)
    description = Column(String, nullable=True)


# TBD: add a variableNumber column to be used in cases like GCAM that aren't matrix oriented?
class InValue(CoreMCSMixin, ORMBase):
    inputId  = Column(Integer, ForeignKey('input.inputId', ondelete="CASCADE"), primary_key=True)
    simId    = Column(Integer, ForeignKey('sim.simId', ondelete="CASCADE"), primary_key=True)
    trialNum = Column(Integer, primary_key=True)
    row      = Column(Integer, primary_key=True)    # TBD: drop?
    col      = Column(Integer, primary_key=True)    # TBD: drop?
    value    = Column(Float)
    __table_args__ = (Index("invalue_index1", "inputId", unique=False),)


class Output(CoreMCSMixin, ORMBase):
    outputId    = Column(Integer, primary_key=True)
    programId   = Column(Integer, ForeignKey('program.programId', ondelete="CASCADE"))
    name        = Column(String)
    timeseries  = Column(Boolean, default=False)    # TBD: use this!
    description = Column(String, nullable=True)

    # TBD: add this for display purposes
    #units       = Column(String, nullable=True)


class OutValue(CoreMCSMixin, ORMBase):
    outputId = Column(Integer, ForeignKey('output.outputId', ondelete="CASCADE"), primary_key=True)
    runId    = Column(Integer, ForeignKey('run.runId', ondelete="CASCADE"), primary_key=True)
    value    = Column(Float)

# deprecated
class Program(CoreMCSMixin, ORMBase):
    programId   = Column(Integer, primary_key=True)
    name        = Column(String)
    description = Column(String, nullable=True)


class Run(CoreMCSMixin, ORMBase):
    runId     = Column(Integer, primary_key=True)
    simId     = Column(Integer, ForeignKey('sim.simId', ondelete="CASCADE"))
    expId     = Column(Integer, ForeignKey('experiment.expId', ondelete="CASCADE"))
    trialNum  = Column(Integer)
    jobNum    = Column(String,   nullable=True)
    queueTime = Column(DateTime, default=datetime.now)
    startTime = Column(DateTime, nullable=True)
    endTime   = Column(DateTime, nullable=True)
    duration  = Column(Integer,  nullable=True)
    status    = Column(String,   nullable=True)
    __table_args__ = (Index("run_index1", "simId", "trialNum", "expId", unique=True),)

class Sim(CoreMCSMixin, ORMBase):
    simId       = Column(Integer, primary_key=True)
    trials      = Column(Integer)
    description = Column(String, nullable=True)
    stamp       = Column(DateTime, default=datetime.now, onupdate=datetime.now)


# Map region numbers to region names. Both ids and names must be unique.
class Region(CoreMCSMixin, ORMBase):
    regionId = Column(Integer, primary_key=True, autoincrement=False)
    canonName = Column(String)        # canonical name for lookup (lowercase, "_" changed to " ")
    displayName = Column(String)        # display name
    __table_args__ = (Index("region_index1", "canonName", unique=True),)


class TimeSeries(CoreMCSMixin, ORMBase):
    '''
    N.B. Columns representing the years of interest are added dynamically as
    "y2005", "y2010", etc.
    '''
    seriesId = Column(Integer, primary_key=True)
    runId = Column(Integer, ForeignKey('run.runId', ondelete="CASCADE"))
    regionId = Column(Integer, ForeignKey('region.regionId', ondelete="CASCADE"))
    outputId = Column(Integer, ForeignKey('output.outputId', ondelete="CASCADE"))
    units = Column(String)
